"""Tools for loading satellite EO data using STAC.

This module provides utilities for loading EO data via
SpatioTemporal Asset Catalog (STAC) metadata, such as
those available on platforms like Microsoft Planetary Computer.
"""

# Used to postpone evaluation of type annotations
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
import odc.stac
import planetary_computer
import pystac_client
import xarray as xr
from odc.geo.geom import BoundingBox
from odc.stac._mdtools import _normalize_geometry

# Only import if running type checking
if TYPE_CHECKING:
    from pystac import ItemCollection
    from xarray import Dataset


def _get_bbox(bbox=None, geopolygon=None, lon=None, lat=None):
    """Obtain an EPSG:4326 bounding box for STAC querying."""
    # If provided as a bounding box, either convert to `odc-geo`
    # bounding box or return as-is
    if bbox is not None:
        bbox_extracted = BoundingBox(*bbox, crs="EPSG:4326") if isinstance(bbox, list | tuple) else bbox

    # If data is provided as a geopolygon, normalise to `odc-geo`
    # geometry and extract bounding box
    elif geopolygon is not None:
        geopolygon = _normalize_geometry(geopolygon)
        bbox_extracted = geopolygon.boundingbox

    # If provided as lon/lat ranges, convert to an `odc-geo` bounding box
    elif (lon is not None) and (lat is not None):
        bbox_extracted = BoundingBox.from_xy(lon, lat, crs="EPSG:4326")

    # Raise error if no valid inputs are provided
    else:
        err_msg = "Must provide both `lon` and `lat`, or `geopolygon`, or `bbox`."
        raise Exception(err_msg)

    # Convert bounding box to EPSG:4326 if required
    if not bbox_extracted.crs.geographic:
        bbox_extracted = bbox_extracted.to_crs("EPSG:4326")

    return bbox_extracted, geopolygon


def stac_load(
    product: str,
    bands: str | list[str] | tuple[str, ...] | None = None,
    time: tuple[str, str] | None = None,
    lon: tuple[float, float] | None = None,
    lat: tuple[float, float] | None = None,
    bbox: tuple[float, float, float, float] | None = None,
    geopolygon: Any | None = None,
    mask_geopolygon: bool = False,
    stac_query: dict | None = None,
    stac_url: str = "https://planetarycomputer.microsoft.com/api/stac/v1",
    dtype: Any | None = None,
    **load_params,
) -> tuple[Dataset, ItemCollection]:
    """Query and load satellite data from a STAC API.

    Uses `pystac_client` to query a SpatioTemporal Asset Catalogue (STAC),
    then load the results as a multi-temporal `xarray.Dataset` using
    `odc-stac`.

    Defaults to using the Microsoft Planetary Computer STAC API.

    Parameters
    ----------
    product : str
        The name of the product (i.e. STAC "collection") to load.
    bands : str or list, optional
        List of band names to load. Defaults to all, also accepts a
        single band name (e.g. "red").
    time : tuple, optional
        The time range to load data for as a tuple of strings (e.g.
        `("2020", "2021")`. If not provided, data will be loaded for
        all available timesteps.
    lon, lat : tuple, optional
        Tuples defining the spatial x and y extent to load in degrees.
    bbox : tuple, optional
        Load data into the extent of a bounding box (left, bottom, right, top).
    geopolygon : multiple types, optional
        Load data into the extents of a geometry. This could be an
        odc.geo Geometry, a GeoJSON dictionary, Shapely geometry, GeoPandas
        DataFrame or GeoSeries. GeoJSON and Shapely inputs are assumed to
        be in EPSG:4326 coordinates.
    mask_geopolygon : bool, optional
        Whether to mask pixels as `NaN` if they are outside the extent
        of a provided geopolygon. Defaults to False; note that this
        will convert all bands to `float32` dtype, so should be used with
        caution for any integer or boolean bands (e.g. cloud masks etc).
    stac_query : dict, optional
        A query dictionary to further filter the data using STAC metadata.
        If not provided, no additional filtering will be applied. For
        example: `stac_query = {"eo:cloud_cover": {"lt": 10}}`.
    stac_url : str, optional
        The URL of the STAC API endpoint to query and load data from.
        Defaults to "https://planetarycomputer.microsoft.com/api/stac/v1".
    dtype : optional
        Data type to load data into. The default will use the dataset's
        default dtype. If `mask_geopolygon=True`, data will be returned
        in `float32` with pixels outside the mask set to `NaN`.
    **load_params : dict
        Additional parameters to be passed to `odc.stac.load()` to customise
        how data is loaded.

    Returns
    -------
    ds : xarray.Dataset
        The loaded dataset as an `xarray.Dataset`.
    items : pystac.item_collection.ItemCollection
        STAC items returned by `pystac_client`.

    """
    # Connect to client
    catalog = pystac_client.Client.open(
        stac_url,
        modifier=(planetary_computer.sign_inplace if "planetarycomputer" in stac_url else None),
    )

    # Set dtype; use provided unless `mask_geopolygon` is provided,
    # in which case use `float32`.
    dtype = "float32" if mask_geopolygon else dtype

    # Set up time for query
    time = "/".join(time) if time is not None else None

    # Extract degree lat/lon bounding box for STAC query
    bbox_4326, geopolygon = _get_bbox(bbox=bbox, geopolygon=geopolygon, lon=lon, lat=lat)

    # Find matching items
    search = catalog.search(
        collections=product,
        bbox=(bbox_4326.left, bbox_4326.bottom, bbox_4326.right, bbox_4326.top),
        datetime=time,
        query=stac_query if stac_query is not None else None,
    )

    # Check how many items were returned
    items = search.item_collection()
    print(f"Found {len(items)} STAC items for {product}")

    # Load with ODC STAC
    ds = odc.stac.load(
        items=items,
        bands=bands,
        bbox=bbox,
        geopolygon=geopolygon,
        lon=lon,
        lat=lat,
        dtype=dtype,
        **load_params,
    )

    # Optionally mask areas outside of supplied geopolygon
    if mask_geopolygon & (geopolygon is not None):
        ds = ds.odc.mask(poly=geopolygon)

    return ds, items


def load_ndwi_mpc(
    time: tuple[str, str] | None = None,
    lon: tuple[float, float] | None = None,
    lat: tuple[float, float] | None = None,
    bbox: BoundingBox | tuple | list | None = None,
    geopolygon: Any | None = None,
    mask_geopolygon: bool = False,
    crs: str = "utm",
    resolution: float = 30,
    resampling: str = "cubic",
    cloud_cover: float = 90,
    load_ls: bool = True,
    load_s2: bool = True,
    chunks: dict[str, int] = {"x": 2048, "y": 2048},
    fail_on_error: bool = False,
) -> Dataset:
    """Load an NDWI time-series from Landsat and/or Sentinel-2 from Microsoft Planetary Computer.

    Landsat and Sentinel-2 satellite data are accessed from the Microsoft
    Planetary Computer ("landsat-c2-l2" and "sentinel-2-l2a") STAC API using
    a set of opinionated loading and cloud-masking parameters.
    These parameters may not be optimal for all use cases; we recommend
    reviewing and/or modifying them prior to formal analysis.

    Parameters
    ----------
    time : tuple, optional
        The time range to load data for as a tuple of strings (e.g.
        `("2020", "2021")`. If not provided, data will be loaded for
        all available timesteps.
    lon, lat : tuple, optional
        Tuples defining the spatial x and y extent to load in degrees.
    bbox : BoundingBox, tuple or list, optional
        Load data into the extent of a bounding box (left, bottom, right, top).
    geopolygon : multiple types, optional
        Load data into the extents of a geometry. This could be an
        odc.geo Geometry, a GeoJSON dictionary, Shapely geometry, GeoPandas
        DataFrame or GeoSeries. GeoJSON and Shapely inputs are assumed to
        be in EPSG:4326 coordinates.
    mask_geopolygon : bool, optional
        Whether to mask pixels as nodata if they are outside the extent
        of a provided geopolygon. Defaults to False.
    crs : str, optional
        The Coordinate Reference System (CRS) to load data into. Defaults
        to "utm", which will attempt to load data into its native UTM
        CRS to minimise resampling.
    resolution : int, optional
        Spatial resolution to load data in. Defaults to 30 metres.
    resampling : str, optional
        Resampling method used for surface reflectance bands. Defaults
        to "cubic"; "nearest" will always be used for categorical cloud
        masking bands.
    cloud_cover : int, optional
        The maximum threshold of cloud cover to load. Defaults to 90%.
    load_ls : bool, optional
        Whether to query and load Landsat data ("landsat-c2-l2").
    load_s2 : bool, optional
        Whether to query and load Sentinel-2 data ("sentinel-2-l2a").
    chunks : dictionary, optional
        Dask chunking used to load data as lazy Dask backed arrays.
        Defaults to `{"x": 2048, "y": 2048}`.
    fail_on_error : bool, optional
        Whether to return an error if any individual satellite datasets
        cannot be loaded. Defaults to False, which prevents ephemeral
        cloud access issues from failing the analysis.

    Returns
    -------
    satellite_ds : xarray.Dataset
        The loaded dataset as an `xarray.Dataset`, containing a single
        "ndwi" `xarray.DataArray`.

    """
    # Assemble parameters used for querying STAC API
    query_params = {
        "time": time,
        "geopolygon": geopolygon,
        "bbox": bbox,
        "lon": lon,
        "lat": lat,
    }

    # Assemble parameters used for loading data into xarray format
    load_params = {
        "crs": crs,
        "resolution": resolution,
        "chunks": chunks,
        "fail_on_error": fail_on_error,
        "groupby": "solar_day",
        "resampling": {"qa_pixel": "nearest", "SCL": "nearest", "*": resampling},
    }

    # List to hold outputs for each sensor (Landsat, Sentinel-2)
    output_list = []

    if load_ls:
        # Load Landsat
        ds_ls, items_ls = stac_load(
            product="landsat-c2-l2",
            bands=("green", "nir08", "qa_pixel"),
            stac_query={
                "eo:cloud_cover": {"lt": cloud_cover},
                "landsat:collection_category": {"in": ["T1"]},
            },
            **query_params,  # type: ignore[arg-type]
            **load_params,  # type: ignore[arg-type]
        )

        # Apply simple Landsat cloud mask
        cloud_mask = (
            # Bit 3: high confidence cloud, bit 4: high confidence shadow
            # https://medium.com/analytics-vidhya/python-for-geosciences-
            # raster-bit-masks-explained-step-by-step-8620ed27141e
            np.bitwise_and(ds_ls.qa_pixel, 1 << 3) | np.bitwise_and(ds_ls.qa_pixel, 1 << 4)
        ) == 0
        ds_ls = ds_ls.where(cloud_mask).drop_vars("qa_pixel")

        # Rescale to between 0.0 and 1.0
        ds_ls = (ds_ls.where(ds_ls != 0) * 0.0000275 + -0.2).clip(0, 1)

        # Convert to NDWI
        ndwi_ls = (ds_ls.green - ds_ls.nir08) / (ds_ls.green + ds_ls.nir08)
        output_list.append(ndwi_ls)

    if load_s2:
        # Load Sentinel-2
        ds_s2, items_s2 = stac_load(
            product="sentinel-2-l2a",
            bands=("green", "nir", "SCL"),
            stac_query={
                "eo:cloud_cover": {"lt": cloud_cover},
            },
            **query_params,  # type: ignore[arg-type]
            **load_params,  # type: ignore[arg-type]
        )

        # Apply simple Sentinel-2 cloud mask
        # 1: defective, 3: shadow, 9: high confidence cloud
        cloud_mask = ~ds_s2.SCL.isin([0, 1, 3, 9])
        ds_s2 = ds_s2.where(cloud_mask).drop_vars("SCL")

        # Sentinel-2 Processing Baseline 4.0 introduced new offset/scaling
        # after January 25 2022. We have to split our data before and after
        # this date, and apply different scaling to each
        ds_s2_pre = ds_s2.sel(time=slice(None, "2022-01-25"))
        ds_s2_post = ds_s2.sel(time=slice("2022-01-26", None))
        ds_s2_pre = (ds_s2_pre.where(ds_s2_pre != 0) * 0.0001).clip(0, 1)
        ds_s2_post = ((ds_s2_post.where(ds_s2_post != 0) - 1000) * 0.0001).clip(0, 1)

        # Combine both rescaled datasets
        ds_s2 = xr.concat([ds_s2_pre, ds_s2_post], dim="time")

        # Convert to NDWI
        ndwi_s2 = (ds_s2.green - ds_s2.nir) / (ds_s2.green + ds_s2.nir)
        output_list.append(ndwi_s2)

    # Merge into a single dataset
    ndwi = xr.concat(output_list, dim="time").sortby("time").to_dataset(name="ndwi")

    # Optionally mask areas outside of supplied geopolygon (this has to be
    # applied here because applying it at the `stac_load` level converts
    # cloud masking bands to "float32".
    if mask_geopolygon & (geopolygon is not None):
        geopolygon = _normalize_geometry(geopolygon)
        ndwi = ndwi.odc.mask(poly=geopolygon)

    return ndwi



"""Tools for analysing local tide dynamics and satellite biases.

This module provides functions to assess how well satellite EO data
captures real-world tides, and reveals potential tide biases in
satellite EO data coverage.
"""

# Used to postpone evaluation of type annotations
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, cast

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from pyTMD.arguments import aliasing_period, frequency

from .eo import _pixel_tides_resample, _resample_chunks, _standardise_inputs, pixel_tides, tag_tides

# Only import if running type checking
if TYPE_CHECKING:
    import os

    from odc.geo.geobox import GeoBox

    from .utils import DatetimeLike


# Satellite revisit times (assuming at equator)
REVISIT_DICT = {
    # SWOT (non-sun-synchronous, KaRIn swath altimetry)
    "swot": 0.99349,
    # Landsat series (sun-synchronous, optical)
    "landsat": 8,  # Combined revisit (e.g. two satallites)
    "landsat-5": 16,
    "landsat-7": 16,
    "landsat-8": 16,
    "landsat-9": 16,
    # Sentinel-2 (sun-synchronous, optical)
    "sentinel-2": 5,  # Combined revisit (e.g. two satallites)
    "sentinel-2a": 10,
    "sentinel-2b": 10,
    "sentinel-2c": 10,
    "sentinel-2d": 10,
    # Sentinel-1 (sun-synchronous, C-band SAR)
    "sentinel-1": 6,  # Combined revisit (e.g. two satallites)
    "sentinel-1a": 12,
    "sentinel-1b": 12,
    "sentinel-1c": 12,
    "sentinel-1d": 12,
    # Sentinel-3 OLCI (sun-synchronous, Ocean and Land Color Instrument)
    "sentinel-3a-olci": 2,
    "sentinel-3b-olci": 2,
    "sentinel-3c-olci": 2,
    "sentinel-3d-olci": 2,
    # Sentinel-3 SLSTR (sun-synchronous, Sea and Land Surface Temperature Radiometer)
    "sentinel-3a-slstr": 1.7,
    "sentinel-3b-slstr": 1.7,
    "sentinel-3c-slstr": 1.7,
    "sentinel-3d-slstr": 1.7,
    # Sentinel-3 SRAL (sun-synchronous, SAR Radar Altimeter)
    "sentinel-3a-sral": 27,
    "sentinel-3b-sral": 27,
    "sentinel-3c-sral": 27,
    "sentinel-3d-sral": 27,
    # NISAR (sun-synchronous, L- and S-band SAR)
    "nisar": 12,
}

# List of major constituents from pyTMD table:
# https://pytmd.readthedocs.io/en/latest/background/Constituent-Table.html
# TODO: read these from code once available in pyTMD package
MAJOR_CONSTITUENTS = {
    "sa": "Solar annual",
    "ssa": "Solar semiannual",
    "mm": "Lunar monthly",
    "msf": "Lunisolar synodic fortnightly",
    "mf": "Lunar declinational fortnightly",
    "mt": "Termensual",
    "2q1": "Smaller elliptical diurnal",
    "sigma1": "Lunar variational diurnal",
    "q1": "Larger lunar elliptical diurnal",
    "rho1": "Larger lunar evectional diurnal",
    "o1": "Lunar diurnal",
    "tau1": "",
    "m1": "Smaller lunar elliptical diurnal",
    "chi1": "Smaller evectional diurnal",
    "pi1": "Solar elliptical diurnal",
    "p1": "Principal solar diurnal",
    "s1": "Raditional solar diurnal",
    "k1": "Principal declinational diurnal",
    "psi1": "Smaller solar elliptical diurnal",
    "phi1": "Second-order solar diurnal",
    "theta1": "Evectional diurnal",
    "j1": "Smaller lunar elliptical diurnal",
    "oo1": "Second-order lunar diurnal",
    "eps2": "",
    "2n2": "Second-order lunar elliptical semidiurnal",
    "mu2": "Lunar variational",
    "n2": "Larger lunar elliptical semidiurnal",
    "nu2": "Larger lunar evectional semidiurnal",
    "m2": "Principal lunar semidiurnal",
    "lambda2": "Smaller lunar evectional",
    "l2": "Smaller lunar elliptical semidiurnal",
    "t2": "Larger solar elliptical semidiurnal",
    "s2": "Principal solar semidiurnal",
    "r2": "Smaller solar elliptical semidiurnal",
    "k2": "Lunisolar declinational semidiurnal",
    "eta2": "",
    "m3": "Principal lunar terdiurnal",
}


def _tide_statistics(
    obs_tides: xr.DataArray,
    all_tides: xr.DataArray,
    min_max_q: tuple = (0.0, 1.0),
    dim: str = "time",
) -> xr.Dataset:
    # Calculate means of observed and modelled tides
    mot = obs_tides.mean(dim=dim)
    mat = all_tides.mean(dim=dim)

    # Identify highest and lowest observed tides
    obs_tides_q = obs_tides.quantile(q=min_max_q, dim=dim).astype("float32")
    lot = obs_tides_q.isel(quantile=0, drop=True)
    hot = obs_tides_q.isel(quantile=-1, drop=True)

    # Identify highest and lowest modelled tides
    all_tides_q = all_tides.quantile(q=min_max_q, dim=dim).astype("float32")
    lat = all_tides_q.isel(quantile=0, drop=True)
    hat = all_tides_q.isel(quantile=-1, drop=True)

    # Calculate tidal range
    otr = hot - lot
    tr = hat - lat

    # Calculate Bishop-Taylor et al. 2018 tidal metrics
    spread = otr / tr
    offset_low_m = lot - lat
    offset_high_m = hat - hot
    offset_low = offset_low_m / tr
    offset_high = offset_high_m / tr

    # Combine into a single dataset
    return xr.merge(
        [
            mot.rename("mot"),
            mat.rename("mat"),
            hot.rename("hot"),
            hat.rename("hat"),
            lot.rename("lot"),
            lat.rename("lat"),
            otr.rename("otr"),
            tr.rename("tr"),
            spread.rename("spread"),
            offset_low.rename("offset_low"),
            offset_high.rename("offset_high"),
        ],
        compat="override",
    )


def _stats_plain_english(
    mot,
    mat,
    hot,
    hat,
    lot,
    lat,
    otr,
    tr,
    spread,
    offset_low,
    offset_high,
) -> None:
    # Plain text descriptors
    mean_diff = "higher" if mot > mat else "lower"
    mean_diff_icon = "⬆️" if mot > mat else "⬇️"
    spread_icon = "🟢" if spread >= 0.9 else "🟡" if 0.7 < spread <= 0.9 else "🔴"
    low_tide_icon = "🟢" if offset_low <= 0.1 else "🟡" if 0.1 <= offset_low < 0.2 else "🔴"
    high_tide_icon = "🟢" if offset_high <= 0.1 else "🟡" if 0.1 <= offset_high < 0.2 else "🔴"

    # Print summary
    print(f"\n\n🌊 Modelled astronomical tide range: {tr:.2f} m ({lat:.2f} to {hat:.2f} m).")
    print(f"🛰️ Observed tide range: {otr:.2f} m ({lot:.2f} to {hot:.2f} m).\n")
    print(f"{spread_icon} {spread:.0%} of the modelled astronomical tide range was observed at this location.")
    print(
        f"{high_tide_icon} The highest {offset_high:.0%} ({offset_high * tr:.2f} m) of the tide range was never observed.",
    )
    print(
        f"{low_tide_icon} The lowest {offset_low:.0%} ({offset_low * tr:.2f} m) of the tide range was never observed.\n",
    )
    print(f"🌊 Mean modelled astronomical tide height: {mat:.2f} m.")
    print(f"🛰️ Mean observed tide height: {mot:.2f} m.")
    print(
        f"{mean_diff_icon} The mean observed tide height was {mot - mat:.2f} m {mean_diff} than the mean modelled astronomical tide height.",
    )


def _stats_figure(
    all_tides_da,
    obs_tides_da,
    hot,
    hat,
    lot,
    lat,
    spread,
    offset_low,
    offset_high,
    plot_var,
    point_col=None,
):
    """Plot tide bias statistics as a figure comparing satellite observations and all modelled tides."""
    # Create plot and add all modelled tides
    fig, ax = plt.subplots(figsize=(10, 6))
    all_tides_da.plot(ax=ax, alpha=0.4, label="Modelled tides")

    # Loop through custom variable values if provided
    if plot_var is not None:
        # Create a list of marker styles
        markers = [
            "o",
            "^",
            "s",
            "D",
            "v",
            "<",
            ">",
            "p",
            "*",
            "h",
            "H",
            "+",
            "x",
            "d",
            "|",
            "_",
        ]

        # Sort values to allow correct grouping
        obs_tides_da = obs_tides_da.sortby("time")
        plot_var = plot_var.sortby("time")

        # Iterate and plot each group
        for i, (label, group) in enumerate(obs_tides_da.groupby(plot_var)):
            group.plot.line(
                ax=ax,
                linewidth=0.0,
                color=point_col,
                marker=markers[i % len(markers)],
                label=label,
                markeredgecolor="black",
                markeredgewidth=0.6,
            )

    # Otherwise, plot all data at once
    else:
        obs_tides_da.plot.line(
            ax=ax,
            marker="o",
            linewidth=0.0,
            color="black" if point_col is None else point_col,
            markeredgecolor="black",
            markeredgewidth=0.6,
            label="Satellite observations",
        )

    # Add legend and remove title
    ax.legend(
        loc="upper center",
        bbox_to_anchor=(0.5, 1.04),
        ncol=20,
        borderaxespad=0,
        frameon=False,
    )
    ax.set_title("")

    # Add horizontal lines for spread/offsets
    ax.axhline(lot, color="black", linestyle=":", linewidth=1)
    ax.axhline(hot, color="black", linestyle=":", linewidth=1)
    ax.axhline(lat, color="black", linestyle=":", linewidth=1)
    ax.axhline(hat, color="black", linestyle=":", linewidth=1)

    # Add text annotations for spread/offsets
    ax.annotate(
        f"    High tide\n    offset ({offset_high:.0%})",
        xy=(all_tides_da.time.max(), np.mean([hat, hot])),
        va="center",
    )
    ax.annotate(
        f"    Spread\n    ({spread:.0%})",
        xy=(all_tides_da.time.max(), np.mean([lot, hot])),
        va="center",
    )
    ax.annotate(
        f"    Low tide\n    offset ({offset_low:.0%})",
        xy=(all_tides_da.time.max(), np.mean([lat, lot])),
    )

    # Remove top right axes and add labels
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.set_ylabel("Tide height (m)")
    ax.set_xlabel("")
    ax.margins(x=0.015)

    return fig


def tide_stats(
    data: xr.Dataset | xr.DataArray | GeoBox,
    time: DatetimeLike | None = None,
    model: str = "EOT20",
    directory: str | os.PathLike | None = None,
    tidepost_lat: float | None = None,
    tidepost_lon: float | None = None,
    plain_english: bool = True,
    plot: bool = True,
    plot_var: str | None = None,
    point_col: str | None = None,
    modelled_freq: str = "3h",
    min_max_q: tuple = (0.0, 1.0),
    round_stats: int = 3,
    **tag_tides_kwargs,
) -> pd.Series:
    """Generate tide statistics and satellite tide bias metrics for every dataset timestep.

    Takes a multi-dimensional dataset and generate tide statistics
    and satellite-observed tide bias metrics, calculated based on
    every timestep in the satellite data and the geographic centroid
    of the imagery.

    By comparing the subset of tides observed by satellites
    against the full astronomical tidal range, we can evaluate
    whether the tides observed by satellites are biased
    (e.g. fail to observe either the highest or lowest tides) due
    to tide aliasing interactions with sun-synchronous satellite
    overpasses.

    For more information about the tidal statistics computed by this
    function, refer to Figure 8 in Bishop-Taylor et al. 2018:
    https://www.sciencedirect.com/science/article/pii/S0272771418308783#fig8

    Parameters
    ----------
    data : xarray.Dataset or xarray.DataArray or odc.geo.geobox.GeoBox
        A multi-dimensional dataset or GeoBox pixel grid that will
        be used to calculate tide statistics. If `data` is an
        xarray object, it should include a "time" dimension.
        If no "time" dimension exists or if `data` is a GeoBox,
        then times must be passed using the `time` parameter.
    time : DatetimeLike, optional
        By default, tides will be modelled using times from the
        "time" dimension of `data`. Alternatively, this param can
        be used to provide a custom set of times. Accepts any format
        that can be converted by `pandas.to_datetime()`. For example:
        `time=pd.date_range(start="2000", end="2001", freq="5h")`
    model : str or list of str, optional
        The tide model (or list of models) to use to model tides.
        If a list is provided, the resulting statistics will be
        returned as a `pandas.Dataframe`; otherwise a `pandas.Series`
        will be returned. Defaults to "EOT20"; specify "all" to use
        all models available in `directory`. For a full list of
        available and supported models, run
        `from eo_tides.utils import list_models; list_models()`.
    directory : str, optional
        The directory containing tide model data files. If no path is
        provided, this will default to the environment variable
        `EO_TIDES_TIDE_MODELS` if set, or raise an error if not.
        Tide modelling files should be stored in sub-folders for each
        model that match the structure required by `pyTMD`
        (<https://geoscienceaustralia.github.io/eo-tides/setup/>).
    tidepost_lat, tidepost_lon : float or int, optional
        Optional coordinates used to model tides. The default is None,
        which uses the centroid of the dataset as the tide modelling
        location.
    plain_english : bool, optional
        An optional boolean indicating whether to print a plain english
        version of the tidal statistics to the screen. Defaults to True;
        only supported when a single tide model is passed to `model`.
    plot : bool, optional
        An optional boolean indicating whether to plot how satellite-
        observed tide heights compare against the full tidal range.
        Defaults to True; only supported when a single tide model is
        passed to `model`.
    plot_var : str, optional
        Optional name of a coordinate, dimension or variable in the array
        that will be used to plot observations with unique symbols.
        Defaults to None, which will plot all observations as circles.
    point_col : str, optional
        Colour used to plot points on the graph. Defaults to None which
        will automatically select colours.
    modelled_freq : str, optional
        An optional string giving the frequency at which to model tides
        when computing the full modelled tidal range. Defaults to '3h',
        which computes a tide height for every three hours across the
        temporal extent of `data`.
    min_max_q : tuple, optional
        Quantiles used to calculate max and min observed and modelled
        astronomical tides. By default `(0.0, 1.0)` which is equivalent
        to minimum and maximum; to use a softer threshold that is more
        robust to outliers, use e.g. `(0.1, 0.9)`.
    round_stats : int, optional
        The number of decimal places used to round the output statistics.
        Defaults to 3.
    **tag_tides_kwargs :
        Optional parameters passed to the `eo_tides.eo.tag_tides`
        function that is used to model tides for each observed and
        modelled timestep.

    Returns
    -------
    stats_df : pandas.Series or pandas.Dataframe
        A pandas object containing the following statistics:

        - `y`: latitude used for modelling tide heights
        - `x`: longitude used for modelling tide heights
        - `mot`: mean tide height observed by the satellite (metres)
        - `mat`: mean modelled astronomical tide height (metres)
        - `lot`: minimum tide height observed by the satellite (metres)
        - `lat`: minimum tide height from modelled astronomical tidal range (metres)
        - `hot`: maximum tide height observed by the satellite (metres)
        - `hat`: maximum tide height from modelled astronomical tidal range (metres)
        - `otr`: tidal range observed by the satellite (metres)
        - `tr`: modelled astronomical tide range (metres)
        - `spread`: proportion of the full modelled tidal range observed by the satellite
        - `offset_low`: proportion of the lowest tides never observed by the satellite
        - `offset_high`: proportion of the highest tides never observed by the satellite

    """
    # Standardise data inputs, time and models
    gbox, obs_times = _standardise_inputs(data, time)

    # Generate range of times covering entire period of satellite record
    assert obs_times is not None  # noqa: S101
    all_times = pd.date_range(
        start=obs_times.min().item(),
        end=obs_times.max().item(),
        freq=modelled_freq,
    )

    # If custom tide modelling locations are not provided, use the
    # dataset centroid
    if not tidepost_lat or not tidepost_lon:
        tidepost_lon, tidepost_lat = gbox.geographic_extent.centroid.coords[0]

    # Model tides for observed timesteps
    obs_tides_da = tag_tides(
        gbox,
        time=obs_times,
        model=model,
        directory=directory,
        tidepost_lat=tidepost_lat,
        tidepost_lon=tidepost_lon,
        **tag_tides_kwargs,
    )

    # Model tides for all modelled timesteps
    all_tides_da = tag_tides(
        gbox,
        time=all_times,
        model=model,
        directory=directory,
        tidepost_lat=tidepost_lat,
        tidepost_lon=tidepost_lon,
        **tag_tides_kwargs,
    )

    # Calculate statistics
    # # (cast ensures typing knows these are always DataArrays)
    stats_ds = _tide_statistics(
        cast("xr.DataArray", obs_tides_da),
        cast("xr.DataArray", all_tides_da),
        min_max_q=min_max_q,
    )

    # Convert to pandas and add tide post coordinates
    stats_df = stats_ds.to_pandas().astype("float32")
    stats_df["x"] = tidepost_lon
    stats_df["y"] = tidepost_lat

    # Convert coordinates to index if dataframe
    if isinstance(stats_df, pd.DataFrame):
        stats_df = stats_df.set_index(["x", "y"], append=True)

    # If a series, print and plot summaries
    else:
        if plain_english:
            _stats_plain_english(
                mot=stats_df.mot,
                mat=stats_df.mat,
                hot=stats_df.hot,
                hat=stats_df.hat,
                lot=stats_df.lot,
                lat=stats_df.lat,
                otr=stats_df.otr,
                tr=stats_df.tr,
                spread=stats_df.spread,
                offset_low=stats_df.offset_low,
                offset_high=stats_df.offset_high,
            )

        if plot:
            _stats_figure(
                all_tides_da=all_tides_da,
                obs_tides_da=obs_tides_da,
                hot=stats_df.hot,
                hat=stats_df.hat,
                lot=stats_df.lot,
                lat=stats_df.lat,
                spread=stats_df.spread,
                offset_low=stats_df.offset_low,
                offset_high=stats_df.offset_high,
                plot_var=data[plot_var] if plot_var else None,
                point_col=point_col,
            )

    # Return in Pandas format
    return stats_df.round(round_stats)


def pixel_stats(
    data: xr.Dataset | xr.DataArray | GeoBox,
    time: DatetimeLike | None = None,
    model: str | list[str] = "EOT20",
    directory: str | os.PathLike | None = None,
    resample: bool = True,
    modelled_freq: str = "3h",
    min_max_q: tuple[float, float] = (0.0, 1.0),
    resample_method: str = "bilinear",
    dask_chunks: tuple[float, float] | None = None,
    dask_compute: bool = True,
    extrapolate: bool = True,
    cutoff: float = 10,
    **pixel_tides_kwargs,
) -> xr.Dataset:
    """Generate tide statistics and satellite tide bias metrics for every dataset pixel.

    Takes a multi-dimensional dataset and generate pixel-level
    tide statistics and satellite-observed tide bias metrics,
    calculated based on every timestep in the satellite data and
    modelled into the spatial extent of the imagery.

    By comparing the subset of tides observed by satellites
    against the full astronomical tidal range, we can evaluate
    whether the tides observed by satellites are biased
    (e.g. fail to observe either the highest or lowest tides)
    due to tide aliasing interactions with sun-synchronous satellite
    overpasses.

    Compared to `tide_stats`, this function models tide metrics
    spatially to produce a two-dimensional output for each statistic.

    For more information about the tidal statistics computed by this
    function, refer to Figure 8 in Bishop-Taylor et al. 2018:
    <https://www.sciencedirect.com/science/article/pii/S0272771418308783#fig8>

    Parameters
    ----------
    data : xarray.Dataset or xarray.DataArray or odc.geo.geobox.GeoBox
        A multi-dimensional dataset or GeoBox pixel grid that will
        be used to calculate spatial tide statistics. If `data`
        is an xarray object, it should include a "time" dimension.
        If no "time" dimension exists or if `data` is a GeoBox,
        then times must be passed using the `time` parameter.
    time : DatetimeLike, optional
        By default, tides will be modelled using times from the
        "time" dimension of `data`. Alternatively, this param can
        be used to provide a custom set of times. Accepts any format
        that can be converted by `pandas.to_datetime()`. For example:
        `time=pd.date_range(start="2000", end="2001", freq="5h")`
    model : str or list of str, optional
        The tide model (or list of models) to use to model tides.
        If a list is provided, a new "tide_model" dimension will be
        added to the `xarray.Dataset` output. Defaults to "EOT20";
        specify "all" to use all models available in `directory`.
        For a full list of available and supported models, run
        `from eo_tides.utils import list_models; list_models()`.
    directory : str, optional
        The directory containing tide model data files. If no path is
        provided, this will default to the environment variable
        `EO_TIDES_TIDE_MODELS` if set, or raise an error if not.
        Tide modelling files should be stored in sub-folders for each
        model that match the structure required by `pyTMD`
        (<https://geoscienceaustralia.github.io/eo-tides/setup/>).
    resample : bool, optional
        Whether to resample tide statistics back into `data`'s original
        higher resolution grid. Set this to `False` if you want to return
        lower-resolution tide statistics (which can be useful for
        assessing tide biases across large spatial extents).
    modelled_freq : str, optional
        An optional string giving the frequency at which to model tides
        when computing the full modelled tidal range. Defaults to '3h',
        which computes a tide height for every three hours across the
        temporal extent of `data`.
    min_max_q : tuple, optional
        Quantiles used to calculate max and min observed and modelled
        astronomical tides. By default `(0.0, 1.0)` which is equivalent
        to minimum and maximum; for a softer threshold that is more
        robust to outliers use e.g. `(0.1, 0.9)`.
    resample_method : str, optional
        If resampling is requested (see `resample` above), use this
        resampling method when resampling from low resolution to high
        resolution pixels. Defaults to "bilinear"; valid options include
        "nearest", "cubic", "min", "max", "average" etc.
    dask_chunks : tuple of float, optional
        Can be used to configure custom Dask chunking for the final
        resampling step. By default, chunks will be automatically set
        to match y/x chunks from `data` if they exist; otherwise chunks
        will be chosen to cover the entire y/x extent of the dataset.
        For custom chunks, provide a tuple in the form `(y, x)`, e.g.
        `(2048, 2048)`.
    dask_compute : bool, optional
        Whether to compute results of the resampling step using Dask.
        If False, `stats_ds` will be returned as a Dask-enabled array.
    extrapolate : bool, optional
        Whether to extrapolate tides into x and y coordinates outside of
        the valid tide modelling domain using nearest-neighbor. Defaults
        to True.
    cutoff : float, optional
        Extrapolation cutoff in kilometers. To avoid producing tide
        statistics too far inland, the default is 10 km.
    **pixel_tides_kwargs :
        Optional parameters passed to the `eo_tides.eo.pixel_tides`
        function.

    Returns
    -------
    stats_ds : xarray.Dataset
        An `xarray.Dataset` containing the following statistics as two-dimensional data variables:

        - `mot`: mean tide height observed by the satellite (metres)
        - `mat`: mean modelled astronomical tide height (metres)
        - `lot`: minimum tide height observed by the satellite (metres)
        - `lat`: minimum tide height from modelled astronomical tidal range (metres)
        - `hot`: maximum tide height observed by the satellite (metres)
        - `hat`: maximum tide height from modelled astronomical tidal range (metres)
        - `otr`: tidal range observed by the satellite (metres)
        - `tr`: modelled astronomical tide range (metres)
        - `spread`: proportion of the full modelled tidal range observed by the satellite
        - `offset_low`: proportion of the lowest tides never observed by the satellite
        - `offset_high`: proportion of the highest tides never observed by the satellite

    """
    # Standardise data inputs, time and models
    gbox, obs_times = _standardise_inputs(data, time)
    dask_chunks = _resample_chunks(data, dask_chunks)
    model = [model] if isinstance(model, str) else model

    # Generate range of times covering entire period of satellite record
    assert obs_times is not None  # noqa: S101
    all_times = pd.date_range(
        start=obs_times.min().item(),
        end=obs_times.max().item(),
        freq=modelled_freq,
    )

    # Model tides for observed timesteps
    obs_tides_da = pixel_tides(
        gbox,
        time=obs_times,
        model=model,
        directory=directory,
        resample=False,
        extrapolate=extrapolate,
        cutoff=cutoff,
        **pixel_tides_kwargs,
    )

    # Model tides for all modelled timesteps
    all_tides_da = pixel_tides(
        gbox,
        time=all_times,
        model=model,
        directory=directory,
        resample=False,
        extrapolate=extrapolate,
        cutoff=cutoff,
        **pixel_tides_kwargs,
    )

    # Calculate statistics
    stats_lowres = _tide_statistics(obs_tides_da, all_tides_da, min_max_q=min_max_q)

    # Assign CRS and geobox to allow reprojection
    stats_lowres = stats_lowres.odc.assign_crs(crs=gbox.crs)

    # Reproject statistics into original high resolution grid
    if resample:
        print("Reprojecting statistics into original resolution")
        return _pixel_tides_resample(
            stats_lowres,
            gbox,
            resample_method,
            dask_chunks,
            dask_compute,
            None,
        )

    print("Returning low resolution statistics array")
    return stats_lowres


def tide_aliasing(
    satellites: list[str] | dict[str, float],
    constituents: list[str] | None = None,
    units: str = "days",
    max_inf: float | None = None,
    style: bool = True,
):
    """Calculate aliasing periods for tidal constituents given satellite revisit intervals.

    This function uses `pyTMD.arguments.aliasing_period` to calculate the
    aliasing periods between satellite overpass periods and the natural
    cycles of tidal constituents. The aliasing period describes how long
    it would take for a satellite to sample the entire tidal cycle for
    each constituent, based on the satellite's observation frequency.

    Short aliasing periods mean the satellite will observe the full range
    of tidal variation relatively quickly, reducing the risk of tide-related
    bias. Long aliasing periods indicate that it will take much longer to
    sample all tidal conditions, increasing the risk that satellite analyses
    may misrepresent tidal dynamics.

    Revisit periods are approximate and based on nominal repeat cycles at the equator.
    Actual observation frequency may vary due to latitude, cloud cover, sensor
    availability, and acquisition planning. Custom revisit intervals can be
    defined by passing a dictionary to `satellites`, e.g. `satellites={"custom-sat": 5}`.

    For more information, refer to https://pytmd.readthedocs.io/en/latest/api_reference/arguments.html#pyTMD.arguments.aliasing_period

    Parameters
    ----------
    satellites : list of str or dict
        List of satellite names to analyse, or a custom dictionary with
        satellite names as keys and revisit frequency in days as values.
        Supported satellites include:

        - Landsat (optical):
            - Two satellites combined: "landsat"
            - Individual: "landsat-5", "landsat-7", "landsat-8", "landsat-9"
        - Sentinel-2 (optical):
            - Two satellites combined: "sentinel-2"
            - Individual: "sentinel-2a", "sentinel-2b", "sentinel-2c"
        - Sentinel-1 (C-band SAR):
            - Two satellites combined: "sentinel-1"
            - Individual: "sentinel-1a", "sentinel-1b", "sentinel-1c"
        - Sentinel-3:
            - OLCI (optical): "sentinel-3a-olci", "sentinel-3b-olci", "sentinel-3c-olci"
            - SLSTR (thermal): "sentinel-3a-slstr", "sentinel-3b-slstr, "sentinel-3c-slstr"
            - SRAL (altimetry): "sentinel-3a-sral", "sentinel-3b-sral, "sentinel-3c-sral"
        - SWOT (KaRIn swath altimetry): "swot"
        - NISAR (L- and S-band SAR): "nisar"
    constituents : list of str or None, optional
        List of tidal constituents to include. If None, use a list of major
        constituents. Constituent names should be lowercase (e.g., "m2", "k1").
    units : str, optional
        Output time units for the aliasing periods. Must be one of:
        "years", "days", "hours", or "minutes". Default is "days".
    max_inf : float, optional
        Maximum aliasing period to display in seconds. Values exceeding
        this threshold are replaced with `np.inf`. Defaults to equivalent of 10
        years if no value is provided.
    style : bool, optional
        If True, returns a styled `pandas.DataFrame`. If False, returns a raw
        DataFrame. Default is True.

    Returns
    -------
    pandas.DataFrame
        A table showing aliasing periods for each tidal constituent across the given satellites.
        The result is styled with a color gradient if `style=True`, or returned as a plain DataFrame.

    Examples
    --------
    >>> eo_tide_aliasing(["sentinel-2", "landsat-8"])
    >>> eo_tide_aliasing(["swot"], constituents=["m2", "k1"], units="hours", style=False)
    >>> eo_tide_aliasing({"custom-sat": 5})

    """
    # If satellites is a dict
    if isinstance(satellites, dict):
        revisit_dict = satellites | REVISIT_DICT
        satellites = list(satellites.keys())
    else:
        revisit_dict = REVISIT_DICT

    # Validate satellite names
    invalid_sats = set(satellites) - set(revisit_dict)
    if invalid_sats:
        valid = ", ".join(sorted(revisit_dict))
        error_msg = f"Unknown satellite(s): {', '.join(invalid_sats)}. Must be one of: {valid}"
        raise ValueError(error_msg)

    # Time unit factors
    unit_factors = {
        "seconds": 1,
        "minutes": 60,
        "hours": 3600,
        "days": 86400,
        "years": 31556952,
    }

    # Use default list of constituents if none provided
    if constituents is None:
        constituents = list(MAJOR_CONSTITUENTS.keys())

    # Extract frequency in radians per second for each constituent,
    # and convert to period in seconds
    omega = np.array([frequency(c)[0] for c in constituents])
    period = 2 * np.pi / omega

    # Compute aliasing period for each satellite
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=RuntimeWarning,
            message="divide by zero encountered in divide",
        )
        aliasing_periods = {}
        for sat in satellites:
            revisit = revisit_dict[sat]
            print(f"Using {revisit} day revisit for {sat}")
            aliasing_periods[("aliasing_period", sat)] = aliasing_period(
                constituents,
                unit_factors["days"] * revisit,
            )

    # Combine into a dataframe
    alias_df = pd.DataFrame(
        index=pd.Index(constituents, name="constituents"),
        data={
            ("period", ""): period,
            **aliasing_periods,
        },
    )

    # Raise error if unit is not supported
    if units not in unit_factors:
        error_msg = f"Unit not supported: {units}; must be one of 'years', 'days', 'hours', 'minutes'"
        raise ValueError(error_msg)

    # Set max value to infinity. If no max is provided, use 10 years
    if max_inf is None:
        max_inf = unit_factors["years"] * 10
    alias_df[alias_df > max_inf] = np.inf

    # Rescale to desired output time units
    precision = 3 if units != "years" else 4
    alias_df = (alias_df / unit_factors[units]).round(precision)

    # Add constituent name column
    alias_df.insert(0, column="name", value=[MAJOR_CONSTITUENTS.get(c) for c in constituents])

    # Style and return
    df_subset = alias_df.loc[:, "aliasing_period"]
    max_col = np.nanquantile(df_subset[np.isfinite(df_subset)].values, 0.9)

    if style:
        return alias_df.style.background_gradient(
            axis=None,
            vmin=0,
            vmax=max_col * 1.5,
            cmap="YlOrRd",
        ).format(precision=precision)
    return alias_df



"""General-purpose utilities for tide model setup and data processing.

This module includes tools for listing and clipping model files,
performing spatial interpolation, and other helper tools used across
the eo_tides package.
"""

# Used to postpone evaluation of type annotations
from __future__ import annotations

import datetime
import os
import pathlib
import textwrap
import warnings
from collections import Counter
from typing import TYPE_CHECKING

# Only import if running type checking
if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, TypeAlias

    from odc.geo.geom import BoundingBox

import numpy as np
import odc.geo
import pandas as pd
import pyTMD
import xarray as xr
from colorama import Style, init
from pyTMD.io.model import load_database
from scipy.spatial import cKDTree as KDTree
from tqdm import tqdm

# Type alias for all possible inputs to "time" params
DatetimeLike: TypeAlias = np.ndarray | pd.DatetimeIndex | pd.Timestamp | datetime.datetime | str | list[str]


def _get_duplicates(array: Sequence[Any]) -> list[Any]:
    """Return any duplicates in a list or array."""
    c = Counter(array)
    return [k for k in c if c[k] > 1]


def _set_directory(
    directory: str | os.PathLike | None = None,
) -> os.PathLike:
    """Set tide modelling files directory.

    If no custom path is provided, try global `EO_TIDES_TIDE_MODELS`
    environmental variable instead.

    For backwards compatibility, also check legacy `DEA_TOOLS_TIDE_MODELS`.
    """
    if directory is None:
        if "EO_TIDES_TIDE_MODELS" in os.environ:
            directory = os.environ["EO_TIDES_TIDE_MODELS"]
        elif "DEA_TOOLS_TIDE_MODELS" in os.environ:
            directory = os.environ["DEA_TOOLS_TIDE_MODELS"]
        else:
            err_msg = (
                "No tide model directory provided via `directory`, and/or no "
                "`EO_TIDES_TIDE_MODELS` environment variable found. "
                "Please provide a valid path to your tide model directory."
            )
            raise Exception(err_msg)

    # Verify path exists
    directory = pathlib.Path(directory).expanduser()
    if not directory.exists():
        err_msg = f"No valid tide model directory found at path `{directory}`"
        raise FileNotFoundError(err_msg)
    return directory


def _standardise_time(
    time: DatetimeLike | None,
) -> np.ndarray | None:
    """Standardise input times for analysis.

    Accept any time format accepted by `pd.to_datetime`, and
    return a datetime64 ndarray. Return None if None passed.
    """
    # Return time as-is if None
    if time is None:
        return None

    # Use pd.to_datetime for conversion, then convert to numpy array
    time = pd.to_datetime(time).to_numpy().astype("datetime64[ns]")

    # Ensure that data has at least one dimension
    return np.atleast_1d(time)


def _standardise_models(
    model: str | list[str],
    directory: str | os.PathLike,
    ensemble_models: list[str] | None = None,
    extra_databases: str | os.PathLike | list | None = None,
) -> tuple[list[str], list[str], list[str] | None]:
    """Standardise lists of models for analysis.

    Take an input model name or list of names, and return a list
    of models to process, requested models, and ensemble models,
    as required by the `model_tides` function.

    Handles two special values passed to `model`: "all", which
    will model tides for all models available in `directory`, and
    "ensemble", which will model tides for all models in a list
    of ensemble models.
    """
    # Turn inputs into arrays for consistent handling
    models_requested = [str(m) for m in np.atleast_1d(model)]

    # Raise error if list contains duplications
    duplicates = _get_duplicates(models_requested)
    if len(duplicates) > 0:
        err_msg = f"The model parameter contains duplicate values: {duplicates}"
        raise ValueError(err_msg)

    # Load supported models from pyTMD database
    available_models, valid_models = list_models(
        directory,
        show_available=False,
        show_supported=False,
        raise_error=True,
        extra_databases=extra_databases,
    )
    custom_options = ["ensemble", "all"]

    # Error if any models are not supported
    if not all(m in valid_models + custom_options for m in models_requested):
        error_text = (
            f"One or more of the requested models are not valid.\n"
            f"Requested models: {models_requested}\n"
            f"Valid models: {valid_models}\n"
            "For tide model setup instructions, refer to the guide: https://geoscienceaustralia.github.io/eo-tides/setup/"
        )
        raise ValueError(error_text) from None

    # Error if any models are not available in `directory`
    if not all(m in available_models + custom_options for m in models_requested):
        error_text = (
            f"One or more of the requested tide models are not available in `{directory}`.\n"
            f"Requested models: {models_requested}\n"
            f"Available models: {available_models}\n"
            "For tide model setup instructions, refer to the guide: https://geoscienceaustralia.github.io/eo-tides/setup/"
        )
        raise ValueError(error_text) from None

    # If "all" models are requested, update requested list to include available models
    if "all" in models_requested:
        models_requested = available_models + [m for m in models_requested if m != "all"]

    # If "ensemble" modeling is requested, use custom list of ensemble models
    if "ensemble" in models_requested:
        print("Running ensemble tide modelling")
        ensemble_models = (
            ensemble_models
            if ensemble_models is not None
            else [
                "EOT20",
                "FES2012",
                "FES2014_extrapolated",
                "FES2022_extrapolated",
                "GOT4.10",
                "GOT5.6_extrapolated",
                "TPXO10-atlas-v2-nc",
                "TPXO8-atlas-nc",
                "TPXO9-atlas-v5-nc",
            ]
        )

        # Error if any ensemble models are not available in `directory`
        if not all(m in available_models for m in ensemble_models):
            error_text = (
                f"One or more of the requested ensemble models are not available in `{directory}`:\n"
                f"{ensemble_models}\n\n"
                f"The following models are available in `{directory}`:\n"
                f"{available_models}"
            )
            raise ValueError(error_text)

        # Return set of all ensemble plus any other requested models
        models_to_process = sorted(set(ensemble_models + [m for m in models_requested if m != "ensemble"]))

    # Otherwise, models to process are the same as those requested
    else:
        models_to_process = models_requested

    return models_to_process, models_requested, ensemble_models


def _clip_model_file(
    nc: xr.Dataset,
    bbox: BoundingBox,
    ydim: str,
    xdim: str,
    ycoord: str,
    xcoord: str,
) -> xr.Dataset:
    """Clips tide model netCDF datasets to a bounding box.

    If the bounding box crosses 0 degrees longitude (e.g. Greenwich prime
    meridian), the dataset will be clipped into two parts and concatenated
    along the x-dimension to create a continuous result.

    Parameters
    ----------
    nc : xr.Dataset
        Input tide model xarray dataset.
    bbox : odc.geo.geom.BoundingBox
        A BoundingBox object for clipping the dataset in EPSG:4326
        degrees coordinates. For example:
        `BoundingBox(left=108, bottom=-48, right=158, top=-6, crs='EPSG:4326')`
    ydim : str
        The name of the xarray dimension representing the y-axis.
        Depending on the tide model, this may or may not contain
        actual latitude values.
    xdim : str
        The name of the xarray dimension representing the x-axis.
        Depending on the tide model, this may or may not contain
        actual longitude values.
    ycoord : str
        The name of the coordinate, variable or dimension containing
        actual latitude values used for clipping the data.
    xcoord : str
        The name of the coordinate, variable or dimension containing
        actual longitude values used for clipping the data.

    Returns
    -------
    xr.Dataset
        A dataset clipped to the specified bounding box, with
        appropriate adjustments if the bounding box crosses 0
        degrees longitude.

    Examples
    --------
    >>> nc = xr.open_dataset("GOT5.5/ocean_tides/2n2.nc")
    >>> bbox = BoundingBox(left=108, bottom=-48, right=158, top=-6, crs="EPSG:4326")
    >>> clipped_nc = _clip_model_file(nc, bbox, xdim="lon", ydim="lat", ycoord="latitude", xcoord="longitude")

    """
    # Extract x and y coords from xarray and load into memory
    xcoords = nc[xcoord].compute()
    ycoords = nc[ycoord].compute()

    # Convert longitudes to 0-360 convention
    left = bbox.left % 360
    right = bbox.right % 360

    # If left coordinate is smaller than right, bbox does not cross
    # zero longitude and can be clipped directly
    if left <= right:  # bbox does not cross 0
        nc_clipped = nc.sel({
            ydim: (ycoords >= bbox.bottom) & (ycoords <= bbox.top),
            xdim: (xcoords >= left) & (xcoords <= right),
        })

    # If left coordinate is larger than right, bbox crosses zero longitude.
    # If so, extract left and right separately and then combine into one
    # concatenated dataset
    elif left > right:  # bbox crosses 0
        # Extract data from left of 0 longitude, and convert lon
        # coords to -180 to 0 range to enable continuous interpolation
        # across 0 boundary
        nc_left = nc.sel({
            ydim: (ycoords >= bbox.bottom) & (ycoords <= bbox.top),
            xdim: (xcoords >= left) & (xcoords <= 360),
        }).assign({xcoord: lambda x: x[xcoord] - 360})

        # Convert additional lon variables for TXPO
        if "lon_v" in nc_left:
            nc_left = nc_left.assign({
                "lon_v": lambda x: x["lon_v"] - 360,
                "lon_u": lambda x: x["lon_u"] - 360,
            })

        # Extract data to right of 0 longitude
        nc_right = nc.sel({
            ydim: (ycoords >= bbox.bottom) & (ycoords <= bbox.top),
            xdim: (xcoords > 0) & (xcoords <= right),
        })

        # Combine left and right data along x dimension
        nc_clipped = xr.concat([nc_left, nc_right], dim=xdim)

        # Temporary fix to remove expanded x dim on lat variables issue
        # for TPXO data; remove x dim by selecting the first obs
        for i in ["lat_z", "lat_v", "lat_u", "con"]:
            try:
                nc_clipped[i] = nc_clipped[i].isel(nx=0)
            except KeyError:  # noqa: PERF203
                pass

    return nc_clipped


def clip_models(
    input_directory: str | os.PathLike,
    output_directory: str | os.PathLike,
    bbox: tuple[float, float, float, float],
    model: list | None = None,
    buffer: float = 5,
    overwrite: bool = False,
) -> None:
    """Clip NetCDF-format ocean tide models to a bounding box.

    This function identifies all NetCDF-format tide models in a
    given input directory, including "ATLAS-netcdf" (e.g. TPXO9-atlas-nc),
    "FES-netcdf" (e.g. FES2022, EOT20), and "GOT-netcdf" (e.g. GOT5.5)
    format files. Files for each model are then clipped to the extent of
    the provided bounding box, handling model-specific file structures.
    After each model is clipped, the result is exported to the output
    directory and verified with `pyTMD` to ensure the clipped data is
    suitable for tide modelling.

    For tide model setup instructions, refer to the guide:
    https://geoscienceaustralia.github.io/eo-tides/setup/

    Parameters
    ----------
    input_directory : str or os.PathLike
        Path to directory containing input NetCDF-format tide model files.
    output_directory : str or os.PathLike
        Path to directory where clipped NetCDF files will be exported.
    bbox : tuple of float
        Bounding box for clipping the tide models in EPSG:4326 degrees
        coordinates, specified as `(left, bottom, right, top)`.
    model : str or list of str, optional
        The tide model (or models) to clip. Defaults to None, which
        will automatically identify and clip all NetCDF-format models
        in the input directly.
    buffer : float, optional
        Buffer distance (in degrees) added to the bounding box to provide
        sufficient data on edges of study area. Defaults to 5 degrees.
    overwrite : bool, optional
        If True, overwrite existing files in the output directory.
        Defaults to False.

    Examples
    --------
    >>> clip_models(
    ...     input_directory="tide_models/",
    ...     output_directory="tide_models_clipped/",
    ...     bbox=(-8.968392, 50.070574, 2.447160, 59.367122),
    ... )

    """
    # Get input and output paths
    input_directory = _set_directory(input_directory)
    output_directory = pathlib.Path(output_directory)

    # Prepare bounding box
    bbox = odc.geo.geom.BoundingBox(*bbox, crs="EPSG:4326").buffered(buffer)

    # Identify NetCDF models
    model_database = load_database()["elevation"]
    netcdf_formats = ["ATLAS-netcdf", "FES-netcdf", "GOT-netcdf"]
    netcdf_models = {k for k, v in model_database.items() if v["format"] in netcdf_formats}

    # Identify subset of available and requested NetCDF models
    available_models, _ = list_models(directory=input_directory, show_available=False, show_supported=False)
    requested_models = list(np.atleast_1d(model)) if model is not None else available_models
    available_netcdf_models = list(set(available_models) & set(requested_models) & set(netcdf_models))

    # Raise error if no valid models found
    if len(available_netcdf_models) == 0:
        err_msg = f"No valid NetCDF models found in {input_directory}."
        raise ValueError(err_msg)

    # If model list is provided,
    print(f"Preparing to clip suitable NetCDF models: {available_netcdf_models}\n")

    # Loop through suitable models and export
    for m in available_netcdf_models:
        # Get model file and grid file list if they exist
        model_files = model_database[m].get("model_file", [])
        grid_file = model_database[m].get("grid_file", [])

        # Convert to list of strings and combine
        model_files = model_files if isinstance(model_files, list) else [model_files]
        grid_file = grid_file if isinstance(grid_file, list) else [grid_file]
        all_files = model_files + grid_file

        # Loop through each model file and clip
        for file in tqdm(all_files, desc=f"Clipping {m}"):
            # Skip if it exists in output directory
            if (output_directory / file).exists() and not overwrite:
                continue

            # Load model file
            nc = xr.open_mfdataset(input_directory / file)

            # Open file and clip according to model
            if m in (
                "GOT5.5",
                "GOT5.5_load",
                "GOT5.5_extrapolated",
                "GOT5.5D",
                "GOT5.5D_extrapolated",
                "GOT5.6",
                "GOT5.6_extrapolated",
            ):
                nc_clipped = _clip_model_file(
                    nc,
                    bbox,
                    xdim="lon",
                    ydim="lat",
                    ycoord="latitude",
                    xcoord="longitude",
                )

            elif m in ("HAMTIDE11",):
                nc_clipped = _clip_model_file(nc, bbox, xdim="LON", ydim="LAT", ycoord="LAT", xcoord="LON")

            elif m in (
                "EOT20",
                "EOT20_load",
                "FES2012",
                "FES2014",
                "FES2014_extrapolated",
                "FES2014_load",
                "FES2022",
                "FES2022_extrapolated",
                "FES2022_load",
            ):
                nc_clipped = _clip_model_file(nc, bbox, xdim="lon", ydim="lat", ycoord="lat", xcoord="lon")

            elif m in (
                "TPXO8-atlas-nc",
                "TPXO9-atlas-nc",
                "TPXO9-atlas-v2-nc",
                "TPXO9-atlas-v3-nc",
                "TPXO9-atlas-v4-nc",
                "TPXO9-atlas-v5-nc",
                "TPXO10-atlas-v2-nc",
            ):
                nc_clipped = _clip_model_file(
                    nc,
                    bbox,
                    xdim="nx",
                    ydim="ny",
                    ycoord="lat_z",
                    xcoord="lon_z",
                )

            else:
                err_msg = f"Model {m} not supported"
                raise Exception(err_msg)

            # Create directory and export
            (output_directory / file).parent.mkdir(parents=True, exist_ok=True)
            nc_clipped.to_netcdf(output_directory / file, mode="w")

        # Verify that models are ready
        if pyTMD.io.model(directory=output_directory).elevation(m=m).verify:
            print(" ✅ Clipped model exported and verified")
        else:
            print(" ❌ Clipped model exported but unable to be verified")

    print(f"\nOutputs exported to {output_directory}")
    list_models(directory=output_directory, show_available=True, show_supported=False)


def list_models(
    directory: str | os.PathLike | None = None,
    show_available: bool = True,
    show_supported: bool = True,
    raise_error: bool = False,
    extra_databases: str | os.PathLike | list | None = None,
) -> tuple[list[str], list[str]]:
    """List all tide models available for tide modelling.

    This function scans the specified tide model directory
    and returns a list of models that are available in the
    directory as well as the full list of all models supported
    by `eo-tides` and `pyTMD`.

    For tide model setup instructions, refer to the guide:
    https://geoscienceaustralia.github.io/eo-tides/setup/

    Parameters
    ----------
    directory : str, optional
        The directory containing tide model data files. If no path is
        provided, this will default to the environment variable
        `EO_TIDES_TIDE_MODELS` if set, or raise an error if not.
        Tide modelling files should be stored in sub-folders for each
        model that match the structure required by `pyTMD`
        (<https://geoscienceaustralia.github.io/eo-tides/setup/>).
    show_available : bool, optional
        Whether to print a list of locally available models.
    show_supported : bool, optional
        Whether to print a list of all supported models, in
        addition to models available locally.
    raise_error : bool, optional
        If True, raise an error if no available models are found.
        If False, raise a warning.
    extra_databases : str or path or list, optional
        Additional custom tide model definitions to load, provided as
        dictionaries or paths to JSON database files. Use this to
        enable custom tide models not included with `pyTMD`.
        See: https://pytmd.readthedocs.io/en/latest/getting_started/Getting-Started.html#model-database

    Returns
    -------
    available_models : list of str
        A list of all tide models available within `directory`.
    supported_models : list of str
        A list of all tide models supported by `eo-tides`.

    """
    init()  # Initialize colorama

    # Set tide modelling files directory. If no custom path is
    # provided, try global environment variable.
    directory = _set_directory(directory)

    # Load supported models from pyTMD database, adding extras if required
    extra_databases = [] if extra_databases is None else extra_databases
    model_database = load_database(extra_databases=extra_databases)["elevation"]

    # Get full list of supported models
    supported_models = list(model_database.keys())

    # Extract expected model paths
    expected_paths = {}
    for m in supported_models:
        model_file = model_database[m]["model_file"]

        # Handle GOT5.6 differently to ensure we test for presence of GOT5.6 constituents
        if m in ("GOT5.6", "GOT5.6_extrapolated"):
            model_file = next(file for file in model_file if "GOT5.6" in file)
        else:
            model_file = model_file[0] if isinstance(model_file, list) else model_file

        # Add expected path to dict, adding directory prefix
        expected_paths[m] = str(directory / pathlib.Path(model_file).expanduser().parent)

    # Define column widths
    status_width = 4  # Width for emoji
    name_width = max(len(name) for name in supported_models)
    path_width = max(len(path) for path in expected_paths.values())

    # Print list of supported models, marking available and
    # unavailable models and appending available to list
    if show_available or show_supported:
        total_width = min(status_width + name_width + path_width + 6, 80)
        print("─" * total_width)
        print(f"{'󠀠🌊':^{status_width}} | {'Model':<{name_width}} | {'Expected path':<{path_width}}")
        print("─" * total_width)

    available_models = []
    for m in supported_models:
        try:
            # Load model
            model_file = pyTMD.io.model(directory=directory, extra_databases=extra_databases).elevation(m=m)

            # Append model to list of available model
            available_models.append(m)

            if show_available:
                # Mark available models with a green tick
                status = "✅"
                print(f"{status:^{status_width}}│ {m:<{name_width}} │ {expected_paths[m]:<{path_width}}")
        except FileNotFoundError:  # noqa: PERF203
            if show_supported:
                # Mark unavailable models with a red cross
                status = "❌"
                print(
                    f"{status:^{status_width}}│ {Style.DIM}{m:<{name_width}} │ {expected_paths[m]:<{path_width}}{Style.RESET_ALL}",
                )

    if show_available or show_supported:
        print("─" * total_width)

        # Print summary
        print(f"\n{Style.BRIGHT}Summary:{Style.RESET_ALL}")
        print(f"Available models: {len(available_models)}/{len(supported_models)}")

    # Raise error or warning if no models are available
    if not available_models:
        warning_msg = textwrap.dedent(
            f"""
            No valid tide models were found in `{directory}`.
            Please ensure that the path you provided is correct, or set the `EO_TIDES_TIDE_MODELS` environment variable to point to a valid tide model directory.
            For tide model setup instructions, refer to the guide: https://geoscienceaustralia.github.io/eo-tides/setup/
            """,
        ).strip()

        if raise_error:
            raise Exception(warning_msg) from None
        warnings.warn(warning_msg, UserWarning, stacklevel=2)

    # Return list of available and supported models
    return available_models, supported_models


def idw(
    input_z,
    input_x,
    input_y,
    output_x,
    output_y,
    p=1,
    k=10,
    max_dist=None,
    k_min=1,
    epsilon=1e-12,
):
    """Perform Inverse Distance Weighting (IDW) interpolation.

    This function performs fast IDW interpolation by creating a KDTree
    from the input coordinates then uses it to find the `k` nearest
    neighbors for each output point. Weights are calculated based on the
    inverse distance to each neighbor, with weights decreasing with
    increasing distance.

    Code inspired by: <https://github.com/DahnJ/REM-xarray>

    Parameters
    ----------
    input_z : array-like
        Array of values at the input points. This can be either a
        1-dimensional array, or a 2-dimensional array where each column
        (axis=1) represents a different set of values to be interpolated.
    input_x : array-like
        Array of x-coordinates of the input points.
    input_y : array-like
        Array of y-coordinates of the input points.
    output_x : array-like
        Array of x-coordinates where the interpolation is to be computed.
    output_y : array-like
        Array of y-coordinates where the interpolation is to be computed.
    p : int or float, optional
        Power function parameter defining how rapidly weightings should
        decrease as distance increases. Higher values of `p` will cause
        weights for distant points to decrease rapidly, resulting in
        nearby points having more influence on predictions. Defaults to 1.
    k : int, optional
        Number of nearest neighbors to use for interpolation. `k=1` is
        equivalent to "nearest" neighbour interpolation. Defaults to 10.
    max_dist : int or float, optional
        Restrict neighbouring points to less than this distance.
        By default, no distance limit is applied.
    k_min : int, optional
        If `max_dist` is provided, some points may end up with less than
        `k` nearest neighbours, potentially producing less reliable
        interpolations. Set `k_min` to set any points with less than
        `k_min` neighbours to NaN. Defaults to 1.
    epsilon : float, optional
        Small value added to distances to prevent division by zero
        errors in the case that output coordinates are identical to
        input coordinates. Defaults to 1e-12.

    Returns
    -------
    interp_values : numpy.ndarray
        Interpolated values at the output coordinates. If `input_z` is
        1-dimensional, `interp_values` will also be 1-dimensional. If
        `input_z` is 2-dimensional, `interp_values` will have the same
        number of rows as `input_z`, with each column (axis=1)
        representing interpolated values for one set of input data.

    Examples
    --------
    >>> input_z = [1, 2, 3, 4, 5]
    >>> input_x = [0, 1, 2, 3, 4]
    >>> input_y = [0, 1, 2, 3, 4]
    >>> output_x = [0.5, 1.5, 2.5]
    >>> output_y = [0.5, 1.5, 2.5]
    >>> idw(input_z, input_x, input_y, output_x, output_y, k=2)
    array([1.5, 2.5, 3.5])

    """
    # Convert to numpy arrays
    input_x = np.atleast_1d(input_x)
    input_y = np.atleast_1d(input_y)
    input_z = np.atleast_1d(input_z)
    output_x = np.atleast_1d(output_x)
    output_y = np.atleast_1d(output_y)

    # Verify input and outputs have matching lengths
    if not (input_z.shape[0] == len(input_x) == len(input_y)):
        err_msg = "All of `input_z`, `input_x` and `input_y` must be the same length."
        raise ValueError(err_msg)
    if len(output_x) != len(output_y):
        err_msg = "Both `output_x` and `output_y` must be the same length."
        raise ValueError(err_msg)

    # Verify k is smaller than total number of points, and non-zero
    if k > input_z.shape[0]:
        err_msg = (
            f"The requested number of nearest neighbours (`k={k}`) "
            f"is smaller than the total number of points ({input_z.shape[0]}).",
        )
        raise ValueError(err_msg)
    if k == 0:
        err_msg = "Interpolation based on `k=0` nearest neighbours is not valid."
        raise ValueError(err_msg)

    # Create KDTree to efficiently find nearest neighbours
    points_xy = np.column_stack((input_y, input_x))
    tree = KDTree(points_xy)

    # Determine nearest neighbours and distances to each
    grid_stacked = np.column_stack((output_y, output_x))
    distances, indices = tree.query(grid_stacked, k=k, workers=-1)

    # If k == 1, add an additional axis for consistency
    if k == 1:
        distances = distances[..., np.newaxis]
        indices = indices[..., np.newaxis]

    # Add small epsilon to distances to prevent division by zero errors
    # if output coordinates are the same as input coordinates
    distances = np.maximum(distances, epsilon)

    # Set distances above max to NaN if specified
    if max_dist is not None:
        distances[distances > max_dist] = np.nan

    # Calculate weights based on distance to k nearest neighbours.
    weights = 1 / np.power(distances, p)
    weights = weights / np.nansum(weights, axis=1).reshape(-1, 1)

    # 1D case: Compute weighted sum of input_z values for each output point
    if input_z.ndim == 1:
        interp_values = np.nansum(weights * input_z[indices], axis=1)

    # 2D case: Compute weighted sum for each set of input_z values
    # weights[..., np.newaxis] adds a dimension for broadcasting
    else:
        interp_values = np.nansum(
            weights[..., np.newaxis] * input_z[indices],
            axis=1,
        )

    # Set any points with less than `k_min` valid weights to NaN
    interp_values[np.isfinite(weights).sum(axis=1) < k_min] = np.nan

    return interp_values



"""eo_tides: Tide modelling tools for large-scale satellite earth observation analysis.

`eo-tides` provides powerful parallelized tools for integrating satellite
Earth observation data with tide modelling. `eo-tides` combines advanced
tide modelling functionality from the `pyTMD` package with `pandas`,
`xarray` and `odc-geo`, providing a suite of flexible tools for efficient
analysis of coastal and ocean Earth observation data – from regional,
continental, to global scale.

These tools can be applied to petabytes of freely available satellite
data (e.g. from Digital Earth Australia or Microsoft Planetary Computer)
loaded via Open Data Cube's `odc-stac` or `datacube` packages, supporting
coastal and ocean earth observation analysis for any time period or
location globally.

Modules
-------
model : Core tide modelling functionality
eo : Tools for integrating satellite EO data with tide modelling
stac : Tools for loading satellite EO data using STAC
stats : Tools for analysing local tide dynamics and satellite biases
utils : General-purpose utilities for tide model setup and data processing
validation : Validation tools for comparing modelled tides to observed tide gauge data.
"""

from importlib.metadata import version

# Import commonly used functions for convenience
from .eo import pixel_tides, tag_tides, tag_timeseries
from .model import ensemble_tides, model_phases, model_tides
from .stac import load_ndwi_mpc, stac_load
from .stats import pixel_stats, tide_aliasing, tide_stats
from .utils import clip_models, idw, list_models
from .validation import eval_metrics, load_gauge_gesla, tide_correlation

# Define what should be imported with "from eo_tides import *"
__all__ = [
    "clip_models",
    "ensemble_tides",
    "eval_metrics",
    "idw",
    "list_models",
    "load_gauge_gesla",
    "load_ndwi_mpc",
    "model_phases",
    "model_tides",
    "pixel_stats",
    "pixel_tides",
    "stac_load",
    "tag_tides",
    "tag_timeseries",
    "tide_aliasing",
    "tide_correlation",
    "tide_stats",
]

# Add version metadata to package
__version__ = version("eo-tides")



"""Validation tools for comparing modelled tides to observed tide gauge data.

This module provides functions for loading, filtering, and analysing
observed tide gauge data to validate modelled tide heights.
"""

import datetime
import os
import warnings
from math import sqrt
from numbers import Number
from pathlib import Path
from typing import cast

import geopandas as gpd
import pandas as pd
import tqdm
import xarray as xr
from odc.geo.geom import BoundingBox, point
from pandas.tseries.offsets import MonthBegin, MonthEnd, YearBegin, YearEnd
from scipy import stats
from shapely.geometry import Point
from sklearn.metrics import mean_absolute_error, mean_squared_error

from .eo import tag_tides
from .stac import load_ndwi_mpc


def eval_metrics(x, y, round=3, all_regress=False):  # noqa: A002
    """Calculate common statistical validation metrics.

    These include:

    * Pearson correlation
    * Root Mean Squared Error
    * Mean Absolute Error
    * R-squared
    * Bias
    * Linear regression parameters (slope, p-value, intercept, standard error)

    Parameters
    ----------
    x : numpy.array
        An array providing "actual" variable values.
    y : numpy.array
        An array providing "predicted" variable values.
    round : int
        Number of decimal places to round each metric
        to. Defaults to 3.
    all_regress : bool
        Whether to return linear regression p-value,
        intercept and standard error (in addition to
        only regression slope). Defaults to False.

    Returns
    -------
    pandas.Series
        A `pd.Series` containing all calculated metrics.

    """
    # Create dataframe to drop na
    xy_df = pd.DataFrame({"x": x, "y": y}).dropna()

    # Compute linear regression
    lin_reg = stats.linregress(x=xy_df.x, y=xy_df.y)

    # Calculate statistics
    stats_dict = {
        "Correlation": xy_df.corr().iloc[0, 1],
        "RMSE": sqrt(mean_squared_error(xy_df.x, xy_df.y)),
        "MAE": mean_absolute_error(xy_df.x, xy_df.y),
        "R-squared": lin_reg.rvalue**2,
        "Bias": (xy_df.y - xy_df.x).mean(),
        "Regression slope": lin_reg.slope,
    }

    # Additional regression params
    if all_regress:
        stats_dict.update({
            "Regression p-value": lin_reg.pvalue,
            "Regression intercept": lin_reg.intercept,
            "Regression standard error": lin_reg.stderr,
        })

    # Return as
    return pd.Series(stats_dict).round(round)


def _round_date_strings(date, round_type="end"):
    """Round a date string up or down to the start or end of a time period.

    Parameters
    ----------
    date : str
        Date string of variable precision (e.g. "2020", "2020-01",
        "2020-01-01").
    round_type : str, optional
        Type of rounding to perform. Valid options are "start" or "end".
        If "start", date is rounded down to the start of the time period.
        If "end", date is rounded up to the end of the time period.
        Default is "end".

    Returns
    -------
    date_rounded : str
        The rounded date string.

    Examples
    --------
    >>> round_date_strings("2020")
    '2020-12-31 00:00:00'

    >>> round_date_strings("2020-01", round_type="start")
    '2020-01-01 00:00:00'

    >>> round_date_strings("2020-01", round_type="end")
    '2020-01-31 00:00:00'

    """
    # Determine precision of input date string
    date_segments = len(date.split("-"))

    # If provided date has no "-", treat it as having year precision
    if date_segments == 1 and round_type == "start":
        date_rounded = str(pd.to_datetime(date) + YearBegin(0))
    elif date_segments == 1 and round_type == "end":
        date_rounded = str(pd.to_datetime(date) + YearEnd(0))

    # If provided date has one "-", treat it as having month precision
    elif date_segments == 2 and round_type == "start":
        date_rounded = str(pd.to_datetime(date) + MonthBegin(0))
    elif date_segments == 2 and round_type == "end":
        date_rounded = str(pd.to_datetime(date) + MonthEnd(0))

    # If more than one "-", then return date as-is
    elif date_segments > 2:
        date_rounded = date

    return date_rounded


def _load_gauge_metadata(metadata_path):
    # Load metadata
    metadata_df = pd.read_csv(metadata_path)
    metadata_df.columns = (
        metadata_df.columns.str.replace(" ", "_", regex=False)
        .str.replace("(", "", regex=False)
        .str.replace(")", "", regex=False)
        .str.replace("/", "_", regex=False)
        .str.lower()
    )
    metadata_df = metadata_df.set_index("site_code")

    # Convert metadata to GeoDataFrame
    metadata_gdf = gpd.GeoDataFrame(
        data=metadata_df,
        geometry=gpd.points_from_xy(metadata_df.longitude, metadata_df.latitude),
        crs="EPSG:4326",
    )

    return metadata_df, metadata_gdf


def _load_gesla_dataset(site, path, na_value):
    # Read dataset
    gesla_df = pd.read_csv(
        path,
        skiprows=41,
        names=["date", "time", "sea_level", "qc_flag", "use_flag"],
        sep=r"\s+",
        na_values=na_value,
    )

    # Combine two date fields
    return (
        gesla_df.assign(
            time=pd.to_datetime(gesla_df["date"] + " " + gesla_df["time"]),
            site_code=site,
        )
        .drop(columns=["date"])
        .set_index("time")
    )


def _nearest_row(gdf, x, y, max_distance=None):
    # Create a point to find the nearest neighbor for
    target_point = gpd.GeoDataFrame({"geometry": [Point(x, y)]}, crs="EPSG:4326")

    # Use sjoin_nearest to find the closest point
    return gpd.sjoin_nearest(target_point, gdf, how="left", max_distance=max_distance)


def load_gauge_gesla(
    x=None,
    y=None,
    site_code=None,
    time=None,
    max_distance=None,
    correct_mean=False,
    filter_use_flag=True,
    site_metadata=True,
    data_path="GESLA4_ALL",
    metadata_path="GESLA4_ALL.csv",
):
    """Load Global Extreme Sea Level Analysis (GESLA) tide gauge data.

    Load and process all available GESLA measured sea-level data
    with an `x, y, time` spatio-temporal query, or from a list of
    specific tide gauges. Can optionally filter by gauge quality
    and append detailed gauge metadata.

    Modified from original code in <https://github.com/philiprt/GeslaDataset>.

    Parameters
    ----------
    x, y : numeric or list/tuple, optional
        Coordinates (in degrees longitude, latitude) used to load GESLA
        tide gauge observations. If provided as singular values
        (e.g. `x=150, y=-32`), then the nearest tide gauge will be returned.
        If provided as a list or tuple (e.g. `x=(150, 152), y=(-32, -30)`),
        then all gauges within the provided bounding box will be loaded.
        Leave as `None` to return all available gauges, or if providing a
        list of site codes using `site_code`.
    site_code : str or list of str, optional
        GESLA site code(s) for which to load data (e.g. `site_code="62650"`).
        If `site_code` is provided, `x` and `y` will be ignored.
    time : tuple or list of str, optional
        Time range to consider, given as a tuple of start and end dates,
        e.g. `time=("2020", "2021")`. The default of None will return all
        tide observations from the year 1800 onward.
    max_distance : numeric, optional
        Optional max distance within which to return the nearest tide gauge
        when `x` and `y` are provided as singular coordinates. Defaults to
        None, which will always return a tide gauge no matter how far away
        it is located from `x` and `y`.
    correct_mean : bool, optional
        Whether to correct sea level measurements to a standardised mean
        sea level by subtracting the mean of all observed sea level
        observations. This can be useful when GESLA tide heights come
        from different or unknown tide datums. Note: the observed mean
        sea level calculated here may differ from true long-term/
        astronomical Mean Sea Level (MSL) datum.
    filter_use_flag : bool, optional
        Whether to filter out low quality observations with a "use_flag"
        value of 0 (do not use). Defaults to True.
    site_metadata : bool, optional
        Whether to add tide gauge station metadata as additional columns
        in the output DataFrame. Defaults to True.
    data_path : str, optional
        Path to the raw GESLA data files ("GESLA-4 DATA", accessible via:
        https://gesla787883612.wordpress.com/downloads/). Defaults to
        "GESLA4_ALL".
    metadata_path : str, optional
        Path to the GESLA station metadata file ("GESLA-4 CSV META-DATA FILE",
        accessible via: https://gesla787883612.wordpress.com/downloads/).
        Defaults to "GESLA4_ALL.csv".

    Returns
    -------
    pd.DataFrame
        Processed GESLA data as a DataFrame with columns including:

        - "time": Timestamps,
        - "sea_level": Observed sea level (m),
        - "qc_flag": Observed sea level QC flag,
        - "use_flag": Use-in-analysis flag (1 = use, 0 = do not use),

        ...and additional columns from station metadata.

    """
    # Expand and validate data and metadata paths
    data_path = Path(data_path).expanduser()
    metadata_path = Path(metadata_path).expanduser()

    if not data_path.exists():
        err_msg = (
            f"GESLA raw data directory not found at `data_path={data_path}`.\n"
            "Download 'GESLA-4 DATA' from: "
            "https://gesla787883612.wordpress.com/downloads/"
        )
        raise FileNotFoundError(err_msg)

    if not metadata_path.exists():
        err_msg = (
            f"GESLA station metadata file not found at: `metadata_path={metadata_path}`.\n"
            "Download the 'GESLA-4 CSV META-DATA FILE' from: "
            "https://gesla787883612.wordpress.com/downloads/"
        )
        raise FileNotFoundError(err_msg)

    # Load tide gauge metadata
    metadata_df, metadata_gdf = _load_gauge_metadata(metadata_path)

    # Use supplied site codes if available
    if site_code is not None:
        site_code = [site_code] if not isinstance(site_code, list) else site_code

    # If x and y are tuples, use xy bounds to identify sites
    elif isinstance(x, tuple | list) & isinstance(y, tuple | list):
        bbox = BoundingBox.from_xy(x, y)
        site_code = metadata_gdf.cx[bbox.left : bbox.right, bbox.top : bbox.bottom].index

    # If x and y are single numbers, select nearest row
    elif isinstance(x, Number) & isinstance(y, Number):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            site_code = (
                _nearest_row(metadata_gdf, x, y, max_distance).rename({"index_right": "site_code"}, axis=1).site_code
            )

        # Raise exception if no valid tide gauges are found
        if site_code.isna().all():
            err_msg = f"No tide gauge found within {max_distance} degrees of {x}, {y}."
            raise Exception(err_msg)

    # Otherwise if all are None, return all available site codes
    elif (site_code is None) & (x is None) & (y is None):
        site_code = metadata_df.index.to_list()

    else:
        err_msg = (
            "`x` and `y` must be provided as either singular coordinates (e.g. `x=150`), or as a tuple bounding box (e.g. `x=(150, 152)`).",
        )
        raise Exception(err_msg)

    # Prepare times
    if time is None:
        time = ["1800", str(datetime.datetime.now().year)]
    time = [time] if not isinstance(time, list | tuple) else time
    start_time = _round_date_strings(time[0], round_type="start")
    end_time = _round_date_strings(time[-1], round_type="end")

    # Identify paths to load and nodata values for each site
    metadata_df["file_name"] = data_path / metadata_df["file_name"]
    paths_na = metadata_df.loc[site_code, ["file_name", "null_value"]]

    # Load and combine into a single dataframe
    gauge_list = [
        _load_gesla_dataset(s, p, na_value=na)
        for s, p, na in tqdm.tqdm(paths_na.itertuples(), total=len(paths_na), desc="Loading GESLA gauges")
    ]
    data_df = pd.concat(gauge_list).sort_index().loc[slice(start_time, end_time)].reset_index().set_index("site_code")

    # Optionally filter by use flag column
    if filter_use_flag:
        data_df = data_df.loc[data_df.use_flag == 1]

    # Optionally insert metadata into dataframe
    if site_metadata:
        if data_df.empty:
            data_df = data_df.reindex(columns=[*data_df.columns, *metadata_df.columns])
        else:
            data_df[metadata_df.columns] = metadata_df.loc[site_code]

    # Add time to index and remove duplicates
    data_df = data_df.set_index("time", append=True)
    duplicates = data_df.index.duplicated()
    if duplicates.sum() > 0:
        warnings.warn("Duplicate timestamps were removed.", UserWarning, stacklevel=2)
        data_df = data_df.loc[~duplicates]

    # Remove observed mean sea level if requested
    if correct_mean:
        data_df["sea_level"] = data_df["sea_level"].sub(data_df.groupby("site_code")["sea_level"].transform("mean"))

    # If no rows are returned, raise a warning
    if data_df.empty:
        warnings.warn(
            f"No data found for site '{site_code}'. "
            "Are you trying to load data using `time` for a period that does not have tide gauge measurements?",
            UserWarning,
            stacklevel=2,
        )

    # Return data
    return data_df


def tide_correlation(
    x: float | None = None,
    y: float | None = None,
    time: tuple[str, str] = ("2022", "2024"),
    crs: str = "EPSG:4326",
    data: xr.DataArray | None = None,
    model: str | list[str] = "all",
    directory: str | os.PathLike | None = None,
    index_threshold: float = 0.0,
    freq_min: float = 0.01,
    freq_max: float = 0.99,
    corr_min: float = 0.15,
    buffer: float = 2500,
    cloud_cover: float = 90,
    load_ls: bool = True,
    load_s2: bool = True,
    **tag_tides_kwargs,
):
    """Ranks tide models based on their correlation with satellite-observed inundation patterns.

    Correlations are calculated between satellite-derived water index
    (e.g. Normalised Difference Water Index, NDWI) and tide heights
    across a buffered region around an input point. High correlations
    indicate that a tide model correctly sorted satellite imagery by
    tide height, with high tide observations being consistently wet,
    and low tide observations being consistently dry.

    By default Microsoft Planetary Computer is used for loading data;
    for advance use, a pre-loaded water index xarray.DataArray from any
    satellite data source can be provided via `data`.

    Parameters
    ----------
    x, y : float, optional
        X and Y coordinates of a coastal point of interest. Assumed
        to be "EPSG:4326" degrees lat/lon; use "crs" for custom CRSs.
    time : tuple, optional
        The time range to load data for as a tuple of strings (e.g.
        `("2022", "2024")`. We recommend using a long enough time
        period (e.g. 3+ years) to ensure that results are not affected
        by tide aliasing; see `eo_tides.stats.tide_aliasing` for more
        information.
    crs : str, optional
        Input coordinate reference system for x and y coordinates.
        Defaults to "EPSG:4326" (WGS84; degrees latitude, longitude).
    data : xr.DataArray, optional
        For advanced use, an xarray.DataArray of water index (e.g. NDWI)
        data can be supplied. If so, all data loading parameters (`x`,
        `y`, `time`, `crs`, `buffer` `cloud_cover`, `load_ls`, `load_s2`)
        will be ignored.
    model : str or list of str, optional
        The tide model (or list of models) to compare. Defaults to
        "all", which will compare all models available in `directory`.
        For a full list of available and supported models,
        run `from eo_tides.utils import list_models; list_models()`.
    directory : str, optional
        The directory containing tide model data files. If no path is
        provided, this will default to the environment variable
        `EO_TIDES_TIDE_MODELS` if set, or raise an error if not.
        Tide modelling files should be stored in sub-folders for each
        model that match the structure required by `pyTMD`
        (<https://geoscienceaustralia.github.io/eo-tides/setup/>).
    index_threshold : float, optional
        Water index (e.g. NDWI) threshold above which a pixel is
        considered wet.
    freq_min : float, optional
        Minimum fraction of time a pixel must be wet to be considered
        intertidal and included in the analysis.
    freq_max : float, optional
        Maximum fraction of time a pixel can be wet to be considered
        intertidal and included in the analysis.
    corr_min : float, optional
        Minimum correlation between water index (e.g. NDWI) and tide
        heights to be considered intertidal and included in the analysis.
        To ensure a like-for-like comparison, model rankings are based on
        the average correlation across every pixel with a positive
        correlation of at least `corr_min` in any individual input model.
    buffer : float, optional
        Radius in meters for generating circular buffer around the
        input point.
    cloud_cover : int, optional
        The maximum threshold of cloud cover to load. Defaults to 90%.
    load_ls : bool, optional
        Whether to load Landsat water index (e.g. NDWI) data.
    load_s2 : bool, optional
        Whether to load Sentinel-2 water index (e.g. NDWI) data.
    **tag_tides_kwargs :
        Optional parameters passed to the `eo_tides.eo.tag_tides`
        function used to model tides.

    Returns
    -------
    corr_df : pandas.DataFrame
        DataFrame with correlation and ranking per tide model.
        Columns include:

        - 'correlation': mean correlation between water index
        (e.g. NDWI and tide heights each model
        - 'rank': model rank based on correlation (1 = highest)
    corr_da : xr.DataArray
        Per-pixel correlations for each model, restricted to likely
        intertidal pixels, with dynamic wetness frequency (e.g. not
        always dry or wet), and with a positive correlation with tide
        heights from at least one tide model.

    Examples
    --------
    >>> from eo_tides import tide_correlation
    >>> y, x = -16.99636, 123.61017
    >>> corr_df, corr_da = tide_correlation(x=x, y=y, directory="tide_models/", cloud_cover=10)

    """
    # Use custom xarray.DataArray if provided
    if data is not None:
        # Verify is xr.DataArray
        if not isinstance(data, xr.DataArray):
            err_msg = "Must provide an xarray.DataArray to `data`."
            raise Exception(err_msg)

        water_index = data
        x, y = data.odc.geobox.geographic_extent.centroid.coords[0]

    # Otherwise load data for point using MPC
    elif (x is not None) and (y is not None):
        # Create circular study area around point
        geom = point(x=x, y=y, crs=crs).to_crs("utm").buffer(buffer).to_crs("EPSG:4326")

        # Load time series water_index (e.g. NDWI) for selected time period and location
        water_index = load_ndwi_mpc(
            time=time,
            geopolygon=geom,
            mask_geopolygon=True,
            cloud_cover=cloud_cover,
            load_ls=load_ls,
            load_s2=load_s2,
        ).ndwi

    # Raise error if no valid inputs are provided
    else:
        err_msg = "Must provide both `x` and `y`, or `data`."
        raise Exception(err_msg)

    # Threshold water_index to identify wet pixels, then calculate
    # overall wetness frequency (making sure NaN pixels are
    # correctly masked to ensure correct statistics)
    wet = (water_index > index_threshold).where(water_index.notnull())
    freq = wet.mean(dim="time")

    # Model tides using selected models (all available by default).
    # Use cast to tell mypy this will always be an xr.DataArray.
    tides_da = cast("xr.DataArray", tag_tides(water_index, model=model, directory=directory, **tag_tides_kwargs))

    # Calculate correlation between wetness and each tide model
    corr = xr.corr(wet, tides_da, dim="time")

    # Restrict data to likely intertidal pixels, with dynamic
    # wetness frequency (e.g. not always dry or wet), and
    # with a positive correlation with tide heights from at
    # least one tide model
    corr_max = corr.max(dim="tide_model")
    corr_da = corr.where((freq >= freq_min) & (freq <= freq_max) & (corr_max >= corr_min))
    corr_da.load()

    # Calculate mean correlation per model and valid data coverage
    corr_mean = corr_da.mean(dim=["x", "y"])
    valid_perc = corr_da.notnull().mean().item()

    # Create DataFrame with correlation and rank
    corr_df = (
        # Convert to dataframe
        corr_mean.to_dataframe(name="correlation")
        .drop("spatial_ref", axis=1)
        # Add rankings
        .assign(rank=lambda df: df.correlation.rank(ascending=False))
        .astype("float32")
        # Add metadata
        .assign(x=x, y=y, valid_perc=valid_perc)
    )

    return corr_df, corr_da



"""Core tide modelling functionality.

This module provides tools for modelling ocean tide heights and phases
for any location or time period using one or more global tide models.
"""

# Used to postpone evaluation of type annotations
from __future__ import annotations

import os
import textwrap
import warnings
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures.process import BrokenProcessPool
from functools import partial
from typing import TYPE_CHECKING

import psutil

# Only import if running type checking
if TYPE_CHECKING:
    import xarray as xr

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import pyTMD
import timescale.time
from tqdm import tqdm

from .utils import (
    DatetimeLike,
    _set_directory,
    _standardise_models,
    _standardise_time,
    idw,
)


def _parallel_splits(
    total_points: int,
    model_count: int,
    parallel_max: int | None = None,
    min_points_per_split: int = 1000,
) -> int:
    """Calculate the optimal number of parallel splits for data processing.

    Optimal parallelisation is estimated based on system resources
    and processing constraints.

    Parameters
    ----------
    total_points : int
        Total number of data points to process
    model_count : int
        Number of models that will be run in parallel
    parallel_max : int, optional
        Maximum number of parallel processes to use. If None, uses CPU core count
    min_points_per_split : int, default=1000
        Minimum number of points that should be processed in each split

    """
    # Get available CPUs. First see if `CPU_GUARANTEE` exists in
    # environment (if running in JupyterHub); if not use psutil
    # followed by standard CPU count
    if parallel_max is None:
        # Take the first valid output
        raw_value = os.environ.get("CPU_GUARANTEE") or psutil.cpu_count(logical=False) or os.cpu_count() or 1

        # Convert to integer
        parallel_max = int(float(raw_value)) if isinstance(raw_value, str) else int(raw_value)

    # Calculate optimal number of splits based on constraints
    splits_by_size = total_points / min_points_per_split
    splits_by_cpu = parallel_max / model_count
    optimal_splits = min(splits_by_size, splits_by_cpu)

    # Convert to integer and ensure at least 1 split
    return int(max(1, optimal_splits))


def _model_tides(
    model,
    x,
    y,
    time,
    directory,
    crs,
    mode,
    output_units,
    method,
    extrapolate,
    cutoff,
    crop,
    crop_buffer,
    append_node,
    constituents,
    extra_databases,
):
    """Worker function applied in parallel by `model_tides`.

    Handles the extraction of tide modelling constituents and tide
    modelling using `pyTMD`.
    """
    # Load models from pyTMD database
    extra_databases = [] if extra_databases is None else extra_databases
    pytmd_model = pyTMD.io.model(
        directory=directory,
        extra_databases=extra_databases,
    ).elevation(model)

    # Reproject x, y to latitude/longitude
    transformer = pyproj.Transformer.from_crs(crs, "EPSG:4326", always_xy=True)
    lon, lat = transformer.transform(x.flatten(), y.flatten())

    # Convert datetime
    ts = timescale.time.Timescale().from_datetime(time.flatten())

    try:
        # Read tidal constants and interpolate to grid points
        amp, ph, c = pytmd_model.extract_constants(
            lon,
            lat,
            type=pytmd_model.type,
            crop=True if crop == "auto" else crop,
            buffer=crop_buffer,
            method=method,
            extrapolate=extrapolate,
            cutoff=cutoff,
            append_node=append_node,
            constituents=constituents,
            extra_databases=extra_databases,
        )

        # TODO: Return constituents
        # print(model, amp.shape)
        # print(amp.shape, ph.shape, c)
        # print(pd.DataFrame({"amplitude": amp}))

    except ValueError:
        # If on-the-fly cropping is auto, try again with crop turned off
        if crop == "auto":
            warnings.warn(
                "On-the-fly cropping is not compatible with the provided "
                "model files; running with `crop=False`. This will not "
                "affect your results but may lead to a minor slowdown. "
                "This can occur when analysing clipped model files restricted "
                "to the western hemisphere. To suppress this warning, manually "
                "set `crop=False`.",
                stacklevel=2,
            )

            # Read tidal constants and interpolate to grid points
            amp, ph, c = pytmd_model.extract_constants(
                lon,
                lat,
                type=pytmd_model.type,
                crop=False,
                buffer=crop_buffer,
                method=method,
                extrapolate=extrapolate,
                cutoff=cutoff,
                append_node=append_node,
                constituents=constituents,
                extra_databases=extra_databases,
            )

        # Otherwise, raise error if cropping if set to True
        else:
            error_msg = (
                "On-the-fly cropping (e.g. `crop=True`) is not compatible with your "
                "provided clipped model files. Please set `crop=False` or `crop='auto'`, "
                "or run your analysis on unclipped global model files to avoid this error."
            )
            raise Exception(error_msg) from None

    # Raise error if constituent files do not cover analysis extent
    except IndexError:
        error_msg = (
            f"The {model} tide model constituent files do not cover the analysis extent "
            f"({min(lon):.2f}, {max(lon):.2f}, {min(lat):.2f}, {max(lat):.2f}). "
            "This can occur if you are using clipped model files to improve run times. "
            "Consider using model files that cover your entire analysis area."
        )
        raise Exception(error_msg) from None

    # Calculate complex phase in radians for Euler's
    cph = -1j * ph * np.pi / 180.0

    # Calculate constituent oscillation
    hc = amp * np.exp(cph)

    # Compute delta times based on model
    deltat = np.zeros_like(ts.tt_ut1) if pytmd_model.corrections in ("OTIS", "ATLAS", "TMD3", "netcdf") else ts.tt_ut1

    # In "one-to-many" mode, extracted tidal constituents and timesteps
    # are repeated/multiplied out to match the number of input points and
    # timesteps, enabling the modeling of tides across all combinations
    # of input times and points. In "one-to-one" mode, no repetition is
    # needed, so each repeat count is set to 1.
    points_repeat = len(x) if mode == "one-to-many" else 1
    time_repeat = len(time) if mode == "one-to-many" else 1
    t, hc, deltat = (
        np.tile(ts.tide, points_repeat),
        hc.repeat(time_repeat, axis=0),
        np.tile(deltat, points_repeat),
    )

    # Create arrays to hold outputs
    tide = np.ma.zeros((len(t)), fill_value=np.nan)
    tide.mask = np.any(hc.mask, axis=1)

    # Predict tidal elevations at time
    tide.data[:] = pyTMD.predict.drift(
        t,
        hc,
        c,
        deltat=deltat,
        corrections=pytmd_model.corrections,
    )

    # Infer minor corrections
    minor = pyTMD.predict.infer_minor(
        t,
        hc,
        c,
        deltat=deltat,
        corrections=pytmd_model.corrections,
        minor=pytmd_model.minor,
    )
    tide.data[:] += minor.data[:]

    # Replace invalid values with fill value
    tide.data[tide.mask] = tide.fill_value

    # Convert data to pandas.DataFrame, and set index to our input
    # time/x/y values
    tide_df = pd.DataFrame(
        {
            "time": np.tile(time, points_repeat),
            "x": np.repeat(x, time_repeat),
            "y": np.repeat(y, time_repeat),
            "tide_model": model,
            "tide_height": tide,
        },
    ).set_index(["time", "x", "y"])

    # Optionally convert outputs to integer units (can save memory)
    if output_units == "m":
        tide_df["tide_height"] = tide_df.tide_height.astype(np.float32)
    elif output_units == "cm":
        tide_df["tide_height"] = (tide_df.tide_height * 100).astype(np.int16)
    elif output_units == "mm":
        tide_df["tide_height"] = (tide_df.tide_height * 1000).astype(np.int16)

    return tide_df


def ensemble_tides(
    tide_df,
    crs,
    ensemble_models,
    ensemble_func=None,
    ensemble_top_n=3,
    ranking_points="https://dea-public-data-dev.s3-ap-southeast-2.amazonaws.com/derivative/dea_intertidal/supplementary/rankings_ensemble_2017-2019.fgb",
    ranking_valid_perc=0.02,
    **idw_kwargs,
):
    """Combine multiple tide models into a single locally optimised ensemble tide model.

    Uses external model ranking data (e.g. satellite altimetry or
    NDWI-tide correlations along the coastline) to inform the
    selection of the best local models.

    This function performs the following steps:

    1. Takes a dataframe of tide heights from multiple tide models, as
       produced by `eo_tides.model.model_tides`
    2. Loads model ranking points from an external file, filters them
       based on the valid data percentage, and retains relevant columns
    3. Interpolates the model rankings into the coordinates of the
       original dataframe using Inverse Weighted Interpolation (IDW)
    4. Uses rankings to combine multiple tide models into a single
       optimised ensemble model (by default, by taking the mean of the
       top 3 ranked models)
    5. Returns a new dataframe with the combined ensemble model predictions

    Parameters
    ----------
    tide_df : pandas.DataFrame
        DataFrame produced by `eo_tides.model.model_tides`, containing
        tide model predictions in long format with columns:
        `["time", "x", "y", "tide_height", "tide_model"]`.
    crs : string
        Coordinate reference system for the "x" and "y" coordinates in
        `tide_df`. Used to ensure that interpolations are performed
        in the correct CRS.
    ensemble_models : list
        A list of models to include in the ensemble modelling process.
        All values must exist as columns with the prefix "rank_" in
        `ranking_points`.
    ensemble_func : dict, optional
        By default, a simple ensemble model will be calculated by taking
        the mean of the `ensemble_top_n` tide models at each location.
        However, a dictionary containing more complex ensemble
        calculations can also be provided. Dictionary keys are used
        to name output ensemble models; functions should take a column
        named "rank" and convert it to a weighting, e.g.:
        `ensemble_func = {"ensemble-custom": lambda x: x["rank"] <= 3}`
    ensemble_top_n : int, optional
        If `ensemble_func` is None, this sets the number of top models
        to include in the mean ensemble calculation. Defaults to 3.
    ranking_points : str, optional
        Path to the file containing model ranking points. This dataset
        should include columns containing rankings for each tide
        model, named with the prefix "rank_". e.g. "rank_EOT20".
        Low values should represent high rankings (e.g. 1 = top ranked).
        The default value points to an example file covering Australia.
    ranking_valid_perc : float, optional
        Minimum percentage of valid data required to include a model
        rank point in the analysis, as defined in a column named
        "valid_perc". Defaults to 0.02.
    **idw_kwargs
        Optional keyword arguments to pass to the `idw` function used
        for interpolation. Useful values include `k` (number of nearest
        neighbours to use in interpolation), `max_dist` (maximum
        distance to nearest neighbours), and `k_min` (minimum number of
        neighbours required after `max_dist` is applied).

    Returns
    -------
    pandas.DataFrame
        DataFrame containing the ensemble model predictions, matching
        the format of the input `tide_df` (e.g. columns `["time", "x",
        "y", "tide_height", "tide_model"]`. By default the 'tide_model'
        column will be labeled "ensemble" for the combined model
        predictions (but if a custom dictionary of ensemble functions is
        provided via `ensemble_func`, each ensemble will be named using
        the provided dictionary keys).

    """
    # Raise data if `tide_df` provided in wide format
    if "tide_model" not in tide_df:
        err_msg = (
            "`tide_df` does not contain the expected 'tide_model' and "
            "'tide_height' columns. Ensure that tides were modelled in "
            "long format (i.e. `output_format='long'` in `model_tides`).",
        )
        raise Exception(err_msg)

    # Extract x and y coords from dataframe
    x = tide_df.index.get_level_values(level="x")
    y = tide_df.index.get_level_values(level="y")

    # Identify input datatype
    input_dtype = tide_df.tide_height.dtype

    # Load model ranks points and reproject to same CRS as x and y
    model_ranking_cols = [f"rank_{m}" for m in ensemble_models]
    try:
        model_ranks_gdf = (
            gpd.read_file(ranking_points, engine="pyogrio")
            .to_crs(crs)
            .query(f"valid_perc > {ranking_valid_perc}")
            .dropna(how="all")
            .filter(model_ranking_cols + ["geometry"])  # noqa: RUF005
        )
    except KeyError:
        error_msg = f"""
        Not all of the expected "rank_" columns {model_ranking_cols} were
        found in the columns of the ranking points file ({ranking_points}).
        Consider passing a custom list of models using `ensemble_models`.
        """
        raise Exception(textwrap.dedent(error_msg).strip()) from None

    # Use points to interpolate model rankings into requested x and y
    id_kwargs_str = "" if idw_kwargs == {} else idw_kwargs
    print(f"Interpolating model rankings using IDW interpolation {id_kwargs_str}")
    ensemble_ranks_df = (
        # Run IDW interpolation on subset of ranking columns
        pd.DataFrame(
            idw(
                input_z=model_ranks_gdf[model_ranking_cols],
                input_x=model_ranks_gdf.geometry.x,
                input_y=model_ranks_gdf.geometry.y,
                output_x=x,
                output_y=y,
                **idw_kwargs,
            ),
            columns=model_ranking_cols,
        )
        .assign(x=x, y=y)
        # Drop any duplicates then melt columns into long format
        .drop_duplicates()
        .melt(id_vars=["x", "y"], var_name="tide_model", value_name="rank")
        # Remove "rank_" prefix to get plain model names
        .replace({"^rank_": ""}, regex=True)
        # Set index columns and rank across groups
        .set_index(["tide_model", "x", "y"])
        .groupby(["x", "y"])
        .rank()
        .astype("float32")  # use smaller dtype for rankings to save memory
    )

    # If no custom ensemble funcs are provided, use a default ensemble
    # calculation that takes the mean of the top N tide models
    if ensemble_func is None:
        ensemble_func = {"ensemble": lambda x: x["rank"] <= ensemble_top_n}

    # Create output list to hold computed ensemble model outputs
    ensemble_list = []

    # Loop through all provided ensemble generation functions
    for ensemble_n, ensemble_f in ensemble_func.items():
        print(f"Combining models into single {ensemble_n} model")

        # Join ranks to input tide data, compute weightings and group
        grouped = (
            # Add tide model as an index so we can join with model ranks
            tide_df.set_index("tide_model", append=True)
            .join(ensemble_ranks_df)
            # Add temp columns containing weightings and weighted values
            .assign(
                weights=ensemble_f,  # use custom func to compute weights
                weighted=lambda i: i.tide_height * i.weights,
            )
            # Groupby is specified in a weird order here as this seems
            # to be the easiest way to preserve correct index sorting
            .groupby(["x", "y", "time"])
        )

        # Use weightings to combine multiple models into single ensemble
        ensemble_df = (
            # Calculate weighted mean
            grouped.weighted.sum()
            .div(grouped.weights.sum())
            # Make sure datatype is the same as the input
            .astype(input_dtype)
            # Convert to dataframe
            .to_frame("tide_height")
            # Label ensemble model and ensure indexes are in expected order
            .assign(tide_model=ensemble_n)
            .reorder_levels(["time", "x", "y"], axis=0)
        )

        ensemble_list.append(ensemble_df)

    # Combine all ensemble models and return as a single dataframe
    return pd.concat(ensemble_list)


def model_tides(
    x: float | list[float] | xr.DataArray,
    y: float | list[float] | xr.DataArray,
    time: DatetimeLike,
    model: str | list[str] = "EOT20",
    directory: str | os.PathLike | None = None,
    crs: str = "EPSG:4326",
    mode: str = "one-to-many",
    output_format: str = "long",
    output_units: str = "m",
    method: str = "linear",
    extrapolate: bool = True,
    cutoff: float | None = None,
    crop: bool | str = "auto",
    crop_buffer: float | None = 5,
    append_node: bool = False,
    constituents: list[str] | None = None,
    parallel: bool = True,
    parallel_splits: int | str = "auto",
    parallel_max: int | None = None,
    ensemble_models: list[str] | None = None,
    extra_databases: str | os.PathLike | list | None = None,
    **ensemble_kwargs,
) -> pd.DataFrame:
    """Model tide heights at multiple coordinates or timesteps using using multiple ocean tide models.

    This function is parallelised to improve performance, and
    supports all tidal models supported by `pyTMD`, including:

    - Empirical Ocean Tide model (EOT20)
    - Finite Element Solution tide models (FES2022, FES2014, FES2012)
    - TOPEX/POSEIDON global tide models (TPXO10, TPXO9, TPXO8)
    - Global Ocean Tide models (GOT5.6, GOT5.5, GOT4.10, GOT4.8, GOT4.7)
    - Hamburg direct data Assimilation Methods for Tides models (HAMTIDE11)
    - Technical University of Denmark tide models (DTU23)

    This function requires access to tide model data files.
    For tide model setup instructions, refer to the guide:
    https://geoscienceaustralia.github.io/eo-tides/setup/

    This function is a modification of the `pyTMD` package's
    `pyTMD.compute.tide_elevations` function. For more info:
    https://pytmd.readthedocs.io/en/latest/api_reference/compute.html#pyTMD.compute.tide_elevations
    https://pytmd.readthedocs.io/en/latest/getting_started/Getting-Started.html#directories

    Parameters
    ----------
    x : float or list of floats
        One or more x coordinates at which to model tides. Assumes
        degrees longitude (EPSG:4326) by default; use `crs` to specify
        a different coordinate reference system.
    y : float or list of floats
        One or more y coordinates at which to model tides. Assumes
        degrees latitude (EPSG:4326) by default; use `crs` to specify
        a different coordinate reference system.
    time : DatetimeLike
        One or more UTC times at which to model tide heights. Accepts
        any time format compatible with `pandas.to_datetime()`, e.g.
        datetime.datetime, pd.Timestamp, pd.DatetimeIndex, numpy.datetime64,
        or date/time strings (e.g. "2020-01-01 23:00"). For example:
        `time = pd.date_range(start="2000", end="2001", freq="5h")`.
    model : str or list of str, optional
        The tide model (or list of models) to use to model tides.
        Defaults to "EOT20"; specify "all" to use all models available
        in `directory`. For a full list of available and supported models,
        run `from eo_tides.utils import list_models; list_models()`.
    directory : str, optional
        The directory containing tide model data files. If no path is
        provided, this will default to the environment variable
        `EO_TIDES_TIDE_MODELS` if set, or raise an error if not.
        Tide modelling files should be stored in sub-folders for each
        model that match the structure required by `pyTMD`
        (<https://geoscienceaustralia.github.io/eo-tides/setup/>).
    crs : str, optional
        Input coordinate reference system for x/y coordinates.
        Defaults to "EPSG:4326" (degrees latitude, longitude).
    mode : str, optional
        Tide modelling analysis mode. Supports two options:

        - `"one-to-many"`: Models tides at every x/y coordinate for
        every timestep in `time`. This is useful for Earth observation
        workflows where you want to model tides at many spatial points
        for a common set of acquisition times (e.g. satellite overpasses).

        - `"one-to-one"`: Model tides using one timestep for each x/y
        coordinate. In this mode, the number of x/y coordinates must
        match the number of timesteps in `time`.
    output_format : str, optional
        Whether to return the output dataframe in long format (with
        results stacked vertically along "tide_model" and "tide_height"
        columns), or wide format (with a column for each tide model).
        Defaults to "long".
    output_units : str, optional
        Units for the returned tide heights. Options are:

        - `"m"` (default): floating point values in metres
        - `"cm"`: integer values in centimetres (x100)
        - `"mm"`: integer values in millimetres (x1000)

        Using integer units can help reduce memory usage.
    method : str, optional
        Method used to interpolate tide model constituent files.
        Defaults to "linear"; options include:

        - `"linear"`, `"nearest"`: scipy regular grid interpolations
        - `"spline"`: scipy bivariate spline interpolation
        - `"bilinear"`: quick bilinear interpolation
    extrapolate : bool, optional
        If True (default), extrapolate tides inland of the valid tide
        model extent using nearest-neighbor interpolation. This can
        ensure tide are returned everywhere, but accuracy may degrade
        with distance from the valid model extent (e.g. inland or along
        complex estuaries or rivers). Set `cutoff` to define the
        maximum extrapolation distance.
    cutoff : float, optional
        Maximum distance in kilometres to extrapolate tides inland of the
        valid tide model extent. The default of None allows extrapolation
        at any (i.e. infinite) distance.
    crop : bool or str, optional
        Whether to crop tide model files on-the-fly to improve performance.
        Defaults to "auto", which enables cropping when supported (some
        clipped model files limited to the western hemisphere may not support
        on-the-fly cropping). Use `crop_buffer` to adjust the buffer
        distance used for cropping.
    crop_buffer : int or float, optional
        The buffer distance in degrees to crop tide model files around the
        requested x/y coordinates. Defaults to 5, which will crop model
        files using a five degree buffer.
    append_node : bool, optional
        Apply adjustments to harmonic constituents to allow for periodic
        modulations over the 18.6-year nodal period (lunar nodal tide).
        Default is False.
    constituents : list, optional
        Optional list of tide constituents to use for tide prediction.
        Default is None, which will use all available constituents.
    parallel : bool, optional
        Whether to parallelise tide modelling. If multiple tide models
        are requested, these will be run in parallel. If enough workers
        are available, the analysis will also be split into spatial
        chunks for additional parallelisation (see "parallel_splits"
        below). Default is True.
    parallel_splits : str or int, optional
        Whether to split the input x and y coordinates into smaller,
        evenly-sized chunks that are processed in parallel. This can
        provide a large performance boost when processing large numbers
        of coordinates. The default is "auto", which will automatically
        attempt to determine optimal splits based on available CPUs,
        the number of input points, and the number of models.
    parallel_max : int, optional
        Maximum number of processes to run in parallel. The default of
        None will automatically determine this from your available CPUs.
    ensemble_models : list of str, optional
        An optional list of models used to generate the ensemble tide
        model if "ensemble" tide modelling is requested. Defaults to
        `["EOT20", "FES2012", "FES2014_extrapolated", "FES2022_extrapolated",
        "GOT4.10", "GOT5.5_extrapolated", "GOT5.6_extrapolated",
        "TPXO10-atlas-v2-nc", "TPXO8-atlas-nc", "TPXO9-atlas-v5-nc"]`.
    extra_databases : str or path or list, optional
        Additional custom tide model definitions to load, provided as
        dictionaries or paths to JSON database files. Use this to
        enable custom tide models not included with `pyTMD`.
        See: https://pytmd.readthedocs.io/en/latest/getting_started/Getting-Started.html#model-database
    **ensemble_kwargs :
        Keyword arguments used to customise the generation of optional
        ensemble tide models if "ensemble" tide modelling is requested.
        These are passed to the underlying `_ensemble_model` function.
        Useful parameters include `ranking_points` (path to model
        rankings data), `k` (for controlling how model rankings are
        interpolated), and `ensemble_top_n` (how many top models to use
        in the ensemble calculation).

    Returns
    -------
    pandas.DataFrame
        A dataframe containing modelled tide heights.

    """
    # Turn inputs into arrays for consistent handling
    x = np.atleast_1d(x)
    y = np.atleast_1d(y)
    time = _standardise_time(time)

    # Validate input arguments
    if time is None:
        err_msg = "Times for modelling tides must be provided via `time`."
        raise ValueError(err_msg)

    if method not in ("bilinear", "spline", "linear", "nearest"):
        err_msg = (
            f"Invalid interpolation method '{method}'. Must be one of 'bilinear', 'spline', 'linear', or 'nearest'."
        )
        raise ValueError(err_msg)

    if output_units not in ("m", "cm", "mm"):
        err_msg = "Output units must be either 'm', 'cm', or 'mm'."
        raise ValueError(err_msg)

    if output_format not in ("long", "wide"):
        err_msg = "Output format must be either 'long' or 'wide'."
        raise ValueError(err_msg)

    if not np.issubdtype(x.dtype, np.number):
        err_msg = "`x` must contain only valid numeric values, and must not be None."
        raise TypeError(err_msg)

    if not np.issubdtype(y.dtype, np.number):
        err_msg = "`y` must contain only valid numeric values, and must not be None."
        raise TypeError(err_msg)

    if len(x) != len(y):
        err_msg = "`x` and `y` must be the same length."
        raise ValueError(err_msg)

    if mode == "one-to-one" and len(x) != len(time):
        err_msg = (
            "The number of supplied `x` and `y` points and `time` values must be "
            "identical in 'one-to-one' mode. Use 'one-to-many' mode if you intended "
            "to model multiple timesteps at each point."
        )
        raise ValueError(err_msg)

    # Set tide modelling files directory. If no custom path is
    # provided, try global environment variable.
    directory = _set_directory(directory)

    # Standardise model list, handling "all" and "ensemble" functionality,
    # and any custom tide model definitions
    models_to_process, models_requested, ensemble_models = _standardise_models(
        model=model,
        directory=directory,
        ensemble_models=ensemble_models,
        extra_databases=extra_databases,
    )

    # Update tide modelling func to add default keyword arguments that
    # are used for every iteration during parallel processing
    iter_func = partial(
        _model_tides,
        directory=directory,
        crs=crs,
        mode=mode,
        output_units=output_units,
        method=method,
        extrapolate=extrapolate,
        cutoff=np.inf if cutoff is None else cutoff,
        crop=crop,
        crop_buffer=crop_buffer,
        append_node=append_node,
        constituents=constituents,
        extra_databases=extra_databases,
    )

    # If automatic parallel splits, calculate optimal value
    # based on available parallelisation, number of points
    # and number of models
    if parallel_splits == "auto":
        parallel_splits = _parallel_splits(
            total_points=len(x),
            model_count=len(models_to_process),
            parallel_max=parallel_max,
        )

    # Verify that parallel splits are not larger than number of points
    assert isinstance(parallel_splits, int)  # noqa: S101
    if parallel_splits > len(x):
        err_msg = f"Parallel splits ({parallel_splits}) cannot be larger than the number of points ({len(x)})."
        raise ValueError(err_msg)

    # Parallelise if either multiple models or multiple splits requested
    if parallel & ((len(models_to_process) > 1) | (parallel_splits > 1)):
        with ProcessPoolExecutor(max_workers=parallel_max) as executor:
            print(
                f"Modelling tides with {', '.join(models_to_process)} in parallel (models: {len(models_to_process)}, splits: {parallel_splits})",
            )

            # Optionally split lon/lat points into `splits_n` chunks
            # that will be applied in parallel
            x_split = np.array_split(x, parallel_splits)
            y_split = np.array_split(y, parallel_splits)

            # Get every combination of models and lat/lon points, and
            # extract as iterables that can be passed to `executor.map()`
            # In "one-to-many" mode, pass entire set of timesteps to each
            # parallel iteration by repeating timesteps by number of total
            # parallel iterations. In "one-to-one" mode, split up
            # timesteps into smaller parallel chunks too.
            if mode == "one-to-many":
                model_iters, x_iters, y_iters = zip(
                    *[(m, x_split[i], y_split[i]) for m in models_to_process for i in range(parallel_splits)],
                    strict=False,
                )
                time_iters = [time] * len(model_iters)
            elif mode == "one-to-one":
                time_split = np.array_split(time, parallel_splits)
                model_iters, x_iters, y_iters, time_iters = zip(
                    *[
                        (m, x_split[i], y_split[i], time_split[i])
                        for m in models_to_process
                        for i in range(parallel_splits)
                    ],
                    strict=False,
                )

            # Apply func in parallel, iterating through each input param
            try:
                model_outputs = list(
                    tqdm(
                        executor.map(
                            iter_func,
                            model_iters,
                            x_iters,
                            y_iters,
                            time_iters,
                        ),
                        total=len(model_iters),
                    ),
                )
            except BrokenProcessPool:
                err_msg = (
                    "Parallelised tide modelling failed, likely to to an out-of-memory error. "
                    "Try reducing the size of your analysis, or set `parallel=False`."
                )
                raise RuntimeError(err_msg) from None

    # Model tides in series if parallelisation is off
    else:
        model_outputs = []

        for model_i in models_to_process:
            print(f"Modelling tides with {model_i}")
            tide_df = iter_func(model_i, x, y, time)
            model_outputs.append(tide_df)

    # Combine outputs into a single dataframe
    tide_df = pd.concat(model_outputs, axis=0)

    # Optionally compute ensemble model and add to dataframe
    if "ensemble" in models_requested:
        ensemble_df = ensemble_tides(tide_df, crs, ensemble_models, **ensemble_kwargs)

        # Update requested models with any custom ensemble models, then
        # filter the dataframe to keep only models originally requested
        models_requested = list(
            np.union1d(models_requested, ensemble_df.tide_model.unique()),
        )
        tide_df = pd.concat([tide_df, ensemble_df]).query(
            "tide_model in @models_requested",
        )

    # Optionally convert to a wide format dataframe with a tide model in
    # each dataframe column
    if output_format == "wide":
        # Pivot into wide format with each time model as a column
        print("Converting to a wide format dataframe")
        tide_df = tide_df.pivot(columns="tide_model", values="tide_height")  # noqa: PD010

        # If in 'one-to-one' mode, reindex using our input time/x/y
        # values to ensure the output is sorted the same as our inputs
        if mode == "one-to-one":
            output_indices = pd.MultiIndex.from_arrays(
                [time, x, y],
                names=["time", "x", "y"],
            )
            tide_df = tide_df.reindex(output_indices)

    return tide_df


def model_phases(
    x: float | list[float] | xr.DataArray,
    y: float | list[float] | xr.DataArray,
    time: DatetimeLike,
    model: str | list[str] = "EOT20",
    directory: str | os.PathLike | None = None,
    time_offset: str = "15 min",
    return_tides: bool = False,
    **model_tides_kwargs,
) -> pd.DataFrame:
    """Model tide phases at multiple coordinates or timesteps using multiple ocean tide models.

    Ebb and low phases (low-flow, high-flow, high-ebb, low-ebb)
    are calculated by running the `eo_tides.model.model_tides`
    function twice, once for the requested timesteps, and again
    after subtracting a small time offset (15 mins by default).
    If tides increased over this period, they are assigned as
    "flow"; if they decreased, they are assigned as "ebb".
    Tides are considered "high" if equal or greater than 0
    metres tide height, otherwise "low".

    This function supports all parameters that are supported
    by `model_tides`.

    For tide model setup instructions, refer to the guide:
    https://geoscienceaustralia.github.io/eo-tides/setup/

    Parameters
    ----------
    x : float or list of floats
        One or more x coordinates at which to model tides. Assumes
        degrees longitude (EPSG:4326) by default; use `crs` to specify
        a different coordinate reference system.
    y : float or list of floats
        One or more y coordinates at which to model tides. Assumes
        degrees latitude (EPSG:4326) by default; use `crs` to specify
        a different coordinate reference system.
    time : DatetimeLike
        One or more UTC times at which to model tide heights. Accepts
        any time format compatible with `pandas.to_datetime()`, e.g.
        datetime.datetime, pd.Timestamp, pd.DatetimeIndex, numpy.datetime64,
        or date/time strings (e.g. "2020-01-01 23:00"). For example:
        `time = pd.date_range(start="2000", end="2001", freq="5h")`.
    model : str or list of str, optional
        The tide model (or list of models) to use to model tides.
        Defaults to "EOT20"; specify "all" to use all models available
        in `directory`. For a full list of available and supported models,
        run `from eo_tides.utils import list_models; list_models()`.
    directory : str, optional
        The directory containing tide model data files. If no path is
        provided, this will default to the environment variable
        `EO_TIDES_TIDE_MODELS` if set, or raise an error if not.
        Tide modelling files should be stored in sub-folders for each
        model that match the structure required by `pyTMD`
        (<https://geoscienceaustralia.github.io/eo-tides/setup/>).
    time_offset: str, optional
        The time offset/delta used to generate a time series of
        offset tide heights required for phase calculation. Defaults
        to "15 min"; can be any string passed to `pandas.Timedelta`.
    return_tides: bool, optional
        Whether to return intermediate modelled tide heights as a
        "tide_height" column in the output dataframe. Defaults to False.
    **model_tides_kwargs :
        Optional parameters passed to the `eo_tides.model.model_tides`
        function. Important parameters include `output_format` (e.g.
        whether to return results in wide or long format), `crop`
        (whether to crop tide model constituent files on-the-fly to
        improve performance) etc.

    Returns
    -------
    pandas.DataFrame
        A dataframe containing modelled tide phases.

    """
    # Pop output format and mode for special handling
    output_format = model_tides_kwargs.pop("output_format", "long")
    mode = model_tides_kwargs.pop("mode", "one-to-many")

    # Model tides
    tide_df = model_tides(
        x=x,
        y=y,
        time=time,
        model=model,
        directory=directory,
        **model_tides_kwargs,
    )

    # Model tides for a time 15 minutes prior to each previously
    # modelled satellite acquisition time. This allows us to compare
    # tide heights to see if they are rising or falling.
    pre_df = model_tides(
        x=x,
        y=y,
        time=time - pd.Timedelta(time_offset),
        model=model,
        directory=directory,
        **model_tides_kwargs,
    )

    # Compare tides computed for each timestep. If the previous tide
    # was higher than the current tide, the tide is 'ebbing'. If the
    # previous tide was lower, the tide is 'flowing'
    ebb_flow = (tide_df.tide_height < pre_df.tide_height.to_numpy()).replace(
        {True: "ebb", False: "flow"},
    )

    # If tides are greater than 0, then "high", otherwise "low"
    high_low = (tide_df.tide_height >= 0).replace({True: "high", False: "low"})

    # Combine into one string and add to data
    tide_df["tide_phase"] = high_low.astype(str) + "-" + ebb_flow.astype(str)

    # Optionally convert to a wide format dataframe with a tide model in
    # each dataframe column
    if output_format == "wide":
        # Pivot into wide format with each time model as a column
        print("Converting to a wide format dataframe")
        tide_df = tide_df.pivot(columns="tide_model")  # noqa: PD010

        # If in 'one-to-one' mode, reindex using our input time/x/y
        # values to ensure the output is sorted the same as our inputs
        if mode == "one-to-one":
            output_indices = pd.MultiIndex.from_arrays(
                [time, x, y],
                names=["time", "x", "y"],
            )
            tide_df = tide_df.reindex(output_indices)

        # Optionally drop tides
        if not return_tides:
            return tide_df.drop("tide_height", axis=1)["tide_phase"]

    # Optionally drop tide heights
    if not return_tides:
        return tide_df.drop("tide_height", axis=1)

    return tide_df
